#!/usr/bin/env python3
# Copyright (c) 2020 oatsu
"""
HTSフルラベルをUSTファイルに変換する。
"""
import utaupy as up


def htsnote2ustnote(hts_note: up.hts.Note, d_table: dict = None, joint: str = '') -> up.ust.Note:
    """
    utaupy.hts.Note を utaupy.ust.Note に変換する。

    joint: 音素記号を結合するときにはさむ文字
    """
    ust_note = up.ust.Note()
    # ノート長
    ust_note.length = int(hts_note.length) * 20
    # 音高情報が無かったらC4にする。
    ust_note.notenum = int(str(hts_note.notenum).replace('xx', '60'))
    # テンポ
    ust_note.tempo = hts_note.tempo
    # 拍子情報をラベル部分に書き込む
    ust_note.label = hts_note.beat
    # 歌詞を登録する
    phonemes = [phoneme.identity for phoneme in hts_note.phonemes]
    lyric = joint.join(phonemes).replace('pau', 'R')
    if lyric.endswith('cl') and lyric != 'cl':
        lyric = lyric.replace('cl', ' cl')
    ust_note.lyric = lyric

    return ust_note


def clean_beat(ust):
    """
    不要な拍子情報を削除する
    """
    # 拍子情報を一時的に記憶する変数
    beat = None
    # 拍子情報が変化する部分だけラベルを残す
    for note in ust.notes:
        if note.label == beat:
            del note['Label']
        else:
            beat = note.label


def songobj2ustobj(hts_song: up.hts.Song, d_table: dict = None, joint: str = '') -> up.ust.Ust:
    """
    Song オブジェクトを Ust オブジェクトに変換する。
    joint: 音素記号を結合するときにはさむ文字
    """
    ust = up.ust.Ust()
    # 各ノートを変換する
    for hts_note in hts_song:
        ust_note = htsnote2ustnote(hts_note, d_table, joint=joint)
        ust.notes.append(ust_note)
    # 重複するテンポ情報を削除
    ust.reload_tempo()
    # 重複する拍子情報を削除
    clean_beat(ust)
    return ust


def hts2ust(path_hts, path_ust, path_table=None, joint: str = ''):
    """
    HTSフルコンテキストラベルファイルをUSTファイルに変換する。
    """
    d_table = up.table.load(path_table) if path_table is not None else None
    full_label = up.hts.load(path_hts)
    ust = songobj2ustobj(full_label.song, d_table, joint=joint)
    ust.write(path_ust)


def main():
    """
    ファイル変換をする。
    """
    path_hts = input('path_hts: ')
    path_ust = path_hts.replace('.lab', '_hts2ust.ust')
    path_table = input('path_table: ')
    hts2ust(path_hts, path_ust, path_table)


if __name__ == '__main__':
    main()



#! /usr/bin/env python33
# coding: utf-8
"""
歌唱データベース用のLABファイルとデータを扱うモジュールです。
"""
import logging
from collections import UserList
from typing import List


def main():
    """
    実行されたときの挙動
    """
    print('labファイル読み取り動作テストをします。')
    path_lab = input('path_lab: ')
    label = load(path_lab)
    print(label)


def load_as_plainlist(path, mode='r', encoding='utf-8', time_unit='100ns'):
    """
    labファイルを ふつうの2次元リストとして読み取る。
    旧バージョンの utaupy.label.load() に近い動作をする。
    """
    # labファイルを読み取り
    with open(path, mode=mode, encoding=encoding) as f:
        lines = [s.strip().split() for s in f.readlines()]
    # 入力ファイル末尾の空白行を除去
    while lines[-1] == ['']:
        del lines[-1]

    # リストにする [[開始時刻, 終了時刻, 発音], [], ...]
    if time_unit in ('s', 'sec', 'second'):
        # きりたんDBのモノラベル形式の場合、時刻が 0.0000000[s] なのでfloatを経由する。
        l = [[int(10000000 * float(v[0])), int(10000000 * float(v[1])), v[2]] for v in lines]
    elif time_unit in ('100ns', 'subus', 'subμs'):
        # Sinsyのモノラベル形式の場合、時刻が 1234567[100ns] なのでintにする。
        l = [[int(v[0]), int(v[1]), v[2]] for v in lines]
    else:
        raise ValueError(
            'function argument "time_unit" must be in ["100ns" (recommended), "s"]')
    return l


def load(path, mode='r', encoding='utf-8', time_unit='100ns'):
    """
    labファイルを読み取って Label クラスオブジェクトにする
    時刻を整数にすることに注意
    """
    path = path.strip('"')
    # lab ファイル読み取り
    with open(path, mode=mode, encoding=encoding) as f:
        lines = [s.strip().split(maxsplit=2) for s in f.readlines()]
    # 入力ファイル末尾の空白行を除去
    while lines[-1] == ['']:
        del lines[-1]

    label = Label()
    # label.original_path = path
    # リストにする [[開始時刻, 終了時刻, 発音], [], ...]
    if time_unit in ('s', 'sec', 'second'):
        for v in lines:
            phoneme = Phoneme()
            phoneme.start = int(10000000 * float(v[0]))
            phoneme.end = int(10000000 * float(v[1]))
            phoneme.symbol = v[2]
            label.append(phoneme)

    # Sinsyのモノラベル形式の場合、時刻が 1234567[100ns] なのでintにする。
    elif time_unit in ('100ns', 'subus', 'subμs'):
        for v in lines:
            phoneme = Phoneme()
            phoneme.start = int(v[0])
            phoneme.end = int(v[1])
            phoneme.symbol = v[2]
            label.append(phoneme)
    else:
        raise ValueError(
            'function argument "time_unit" must be in ["100ns" (recommended), "s"]')
    return label


class Label(UserList):
    """
    歌唱ラベルLABファイルを想定したクラス(2019/04/19から)
    """

    # def __init__(self):
    # なんかうまく実装できない
    # ループしようとすると TypeError: __init__() takes 1 positional argument but 2 were given になる
    # super().__init__()
    # self.original_path = None

    def __str__(self):
        """
        文字列として扱うときのフォーマット
        """
        label_as_str = '\n'.join(str(phoneme) for phoneme in self)
        return label_as_str

    @property
    def offset(self) -> int:
        """
        最初の音素の発声までの時間を取得する。
        """
        return self.data[0].start

    @property
    def start_times(self):
        """
        発声開始時刻のリスト
        """
        return [phoneme.start for phoneme in self.data]

    @start_times.setter
    def start_times(self, l: List[int]):
        assert len(l) == self.data
        for phoneme, new_start in zip(self.data, l):
            phoneme.start = new_start

    @property
    def end_times(self):
        """
        発声終了時刻のリスト
        """
        return [phoneme.end for phoneme in self.data]

    @end_times.setter
    def end_times(self, l: List[int]):
        assert len(l) == self.data
        for phoneme, new_end in zip(self.data, l):
            phoneme.end = new_end

    @property
    def contexts(self):
        """
        発声終了時刻のリスト
        """
        return [phoneme.symbol for phoneme in self.data]

    @contexts.setter
    def contexts(self, l: List[int]):
        assert len(l) == self.data
        for phoneme, new_contexts in zip(self.data, l):
            phoneme.symbol = new_contexts

    def shift(self, time_length_100ns: int):
        """
        全体の時刻をずらす。
        """
        for phoneme in self:
            phoneme.start += time_length_100ns
            phoneme.end += time_length_100ns

    def is_valid(self, threshold: int = 0, time_unit='ms') -> bool:
        """
        発声時間が一定未満な音素ラベル行を検出
        threshold: 許容される最小の発声時間(ms)
        """
        if time_unit == 'ms':
            threshold_100ns = int(threshold * (10**4))
        elif time_unit == '100ns':
            threshold_100ns = threshold
        else:
            raise ValueError('Argument "time_unit" must be "ms" or "100ns".')
        # 既定の長さに達しない音素を検出
        invalid_phonemes = []
        for phoneme in self:
            duration = phoneme.end - phoneme.start
            if duration < threshold_100ns:
                invalid_phonemes.append(phoneme)
                logging.error('発声時間が %s%s 未満か負です : %s %s',
                              threshold, time_unit, phoneme.start, phoneme.end)

        # 前後の音素の開始・終了時刻と一致するかチェック
        for i, phoneme in enumerate(self[1:-1], 1):
            previous_phoneme = self[i - 1]
            next_phoneme = self[i + 1]
            if phoneme.start != previous_phoneme.end:
                error_message = '\n'.join([
                    '発声開始時刻が直前の発声終了時刻と一致しません',
                    f'  previous_phoneme    : {previous_phoneme.start} {previous_phoneme.end}',
                    f'  current_phoneme : {phoneme.start} {phoneme.end}'])
                logging.error(error_message)
                invalid_phonemes.append(phoneme)
            if phoneme.end != next_phoneme.start:
                error_message = '\n'.join([
                    '発声終了時刻が直後の発声終了時刻と一致しません',
                    f'  current_phoneme : {phoneme.start} {phoneme.end}',
                    f'  next_phoneme    : {next_phoneme.start} {next_phoneme.end}'])
                logging.error(error_message)
                invalid_phonemes.append(phoneme)

        return len(invalid_phonemes) == 0

    def check_invalid_time(self, threshold=0):
        """
        発声時間が一定未満な音素ラベル行を検出
        threshold: 許容される最小の発声時間(ms)
        """
        threshold_100ns = int(threshold * (10**4))
        for phoneme in self:
            duration = phoneme.end - phoneme.start
            if duration < threshold_100ns:
                error_message = f'発声時間が {threshold}ms 未満か負です'
                logging.error(error_message)
        for i, phoneme in enumerate(self[1:-1], 1):
            previous_phoneme = self[i - 1]
            next_phoneme = self[i + 1]
            if phoneme.start != previous_phoneme.end:
                error_message = '\n'.join([
                    '発声開始時刻が直前の発声終了時刻と一致しません',
                    f'  previous_phoneme: {str(previous_phoneme)}',
                    f'  current_phoneme : {str(phoneme)}'])
                logging.error(error_message)
            if phoneme.end != next_phoneme.start:
                error_message = '\n'.join([
                    '発声終了時刻が直後の発声終了時刻と一致しません',
                    f'  current_phoneme : {str(phoneme)}',
                    f'  next_phoneme    : {str(next_phoneme)}'])

    def reload(self):
        """
        発音開始時刻を参照して、発音終了時刻を自動補完する。
        ただし最終行の発音終了時刻だけは補完できないため、そのままにする。
        """
        for i, phoneme in enumerate(self[:-1]):
            phoneme.end = self[i + 1].start

    def round(self, step_size: int):
        """
        時刻を丸める。
        """
        for phoneme in self:
            phoneme.start = round(phoneme.start / step_size) * step_size
            phoneme.end = round(phoneme.end / step_size) * step_size

    def write(self, path_out, mode='w',
              encoding='utf-8', newline='\n', delimiter=' ', time_unit='100ns'):
        """
        LABファイルを書き出し
        """
        if time_unit == '100ns':
            lines = ['{0}{3}{1}{3}{2}'.format(
                ph.start, ph.end, ph.symbol, delimiter) for ph in self]
        elif time_unit in ('s', '1s', 'sec'):
            lines = ['{:.7f} {:.7f} {}'.format(
                ph.start, ph.end, ph.symbol) for ph in self]  # 100ns -> 1s 表記変換
        else:
            raise ValueError("Argument time_unit must be '100ns' or 's'.")

        # ファイル出力
        with open(path_out, mode=mode, encoding=encoding, newline=newline) as f:
            f.write('\n'.join(lines))
        return lines


class Phoneme:
    """
    ラベルの一行分の情報を持つクラス(2020/07/23から)
    """

    def __init__(self):
        self.start = None   # 発声開始位置
        self.end = None     # 発声終了位置
        self.symbol = None  # 発音記号

    def __str__(self):
        return f'{self.start} {self.end} {self.symbol}'

    @property
    def duration(self) -> int:
        """
        発声時間
        """
        return self.end - self.start


if __name__ == '__main__':
    main()

if __name__ == '__init__':
    pass



#!/usr/bin/env python3
# Copyright (c) 2020 oatsu
"""Python3 module for HTS-full-label.

Sinsy仕様のHTSフルコンテキストラベルを扱うモジュール
forループが多いのでpypy3を使っても良いかも。(とくにdeepcopyが重い)
"""
# import json
import re
from collections import UserList
from copy import copy, deepcopy
from decimal import ROUND_HALF_UP, Decimal
from itertools import chain
from typing import Union

from . import label as _label

# from pprint import pprint

# p1を埋めるのに使う。
VOWELS = ['a', 'i', 'u', 'e', 'o', 'A', 'I', 'U', 'E', 'O', 'N', 'ae', 'AE']
BREAKS = ['br', 'cl']
PAUSES = ['pau']
SILENCES = ['sil']

# e1を埋めるのに使う
NOTENUM_TO_ABSPITCH_DICT = {
    'xx': 'xx',
    '12': 'C0', '13': 'Db0', '14': 'D0', '15': 'Eb0', '16': 'E0', '17': 'F0',
    '18': 'Gb0', '19': 'G0', '20': 'Ab0', '21': 'A0', '22': 'Bb0', '23': 'B0',
    '24': 'C1', '25': 'Db1', '26': 'D1', '27': 'Eb1', '28': 'E1', '29': 'F1',
    '30': 'Gb1', '31': 'G1', '32': 'Ab1', '33': 'A1', '34': 'Bb1', '35': 'B1',
    '36': 'C2', '37': 'Db2', '38': 'D2', '39': 'Eb2', '40': 'E2', '41': 'F2',
    '42': 'Gb2', '43': 'G2', '44': 'Ab2', '45': 'A2', '46': 'Bb2', '47': 'B2',
    '48': 'C3', '49': 'Db3', '50': 'D3', '51': 'Eb3', '52': 'E3', '53': 'F3',
    '54': 'Gb3', '55': 'G3', '56': 'Ab3', '57': 'A3', '58': 'Bb3', '59': 'B3',
    '60': 'C4', '61': 'Db4', '62': 'D4', '63': 'Eb4', '64': 'E4', '65': 'F4',
    '66': 'Gb4', '67': 'G4', '68': 'Ab4', '69': 'A4', '70': 'Bb4', '71': 'B4',
    '72': 'C5', '73': 'Db5', '74': 'D5', '75': 'Eb5', '76': 'E5', '77': 'F5',
    '78': 'Gb5', '79': 'G5', '80': 'Ab5', '81': 'A5', '82': 'Bb5', '83': 'B5',
    '84': 'C6', '85': 'Db6', '86': 'D6', '87': 'Eb6', '88': 'E6', '89': 'F6',
    '90': 'Gb6', '91': 'G6', '92': 'Ab6', '93': 'A6', '94': 'Bb6', '95': 'B6',
    '96': 'C7', '97': 'Db7', '98': 'D7', '99': 'Eb7', '100': 'E7', '101': 'F7',
    '102': 'Gb7', '103': 'G7', '104': 'Ab7', '105': 'A7', '106': 'Bb7', '107': 'B7',
    '108': 'C8', '109': 'Db8', '110': 'D8', '111': 'Eb8', '112': 'E8', '113': 'F8',
    '114': 'Gb8', '115': 'G8', '116': 'Ab8', '117': 'A8', '118': 'Bb8', '119': 'B8',
    '120': 'C9', '121': 'Db9', '122': 'D9', '123': 'Eb9', '124': 'E9', '125': 'F9',
    '126': 'Gb9', '127': 'G9',
}
# e57, e58 を埋めるのに使う
ABSPITCH_TO_NOTENUM_DICT = {
    'xx': 'xx',
    'C0': 12, 'Db0': 13, 'D0': 14, 'Eb0': 15, 'E0': 16, 'F0': 17,
    'Gb0': 18, 'G0': 19, 'Ab0': 20, 'A0': 21, 'Bb0': 22, 'B0': 23,
    'C1': 24, 'Db1': 25, 'D1': 26, 'Eb1': 27, 'E1': 28, 'F1': 29,
    'Gb1': 30, 'G1': 31, 'Ab1': 32, 'A1': 33, 'Bb1': 34, 'B1': 35,
    'C2': 36, 'Db2': 37, 'D2': 38, 'Eb2': 39, 'E2': 40, 'F2': 41,
    'Gb2': 42, 'G2': 43, 'Ab2': 44, 'A2': 45, 'Bb2': 46, 'B2': 47,
    'C3': 48, 'Db3': 49, 'D3': 50, 'Eb3': 51, 'E3': 52, 'F3': 53,
    'Gb3': 54, 'G3': 55, 'Ab3': 56, 'A3': 57, 'Bb3': 58, 'B3': 59,
    'C4': 60, 'Db4': 61, 'D4': 62, 'Eb4': 63, 'E4': 64, 'F4': 65,
    'Gb4': 66, 'G4': 67, 'Ab4': 68, 'A4': 69, 'Bb4': 70, 'B4': 71,
    'C5': 72, 'Db5': 73, 'D5': 74, 'Eb5': 75, 'E5': 76, 'F5': 77,
    'Gb5': 78, 'G5': 79, 'Ab5': 80, 'A5': 81, 'Bb5': 82, 'B5': 83,
    'C6': 84, 'Db6': 85, 'D6': 86, 'Eb6': 87, 'E6': 88, 'F6': 89,
    'Gb6': 90, 'G6': 91, 'Ab6': 92, 'A6': 93, 'Bb6': 94, 'B6': 95,
    'C7': 96, 'Db7': 97, 'D7': 98, 'Eb7': 99, 'E7': 100, 'F7': 101,
    'Gb7': 102, 'G7': 103, 'Ab7': 104, 'A7': 105, 'Bb7': 106, 'B7': 107,
    'C8': 108, 'Db8': 109, 'D8': 110, 'Eb8': 111, 'E8': 112, 'F8': 113,
    'Gb8': 114, 'G8': 115, 'Ab8': 116, 'A8': 117, 'Bb8': 118, 'B8': 119,
    'C9': 120, 'Db9': 121, 'D9': 122, 'Eb9': 123, 'E9': 124, 'F9': 125,
    'Gb9': 126, 'G9': 127
}


def load(source):
    """HTSフルコンテキストラベル(Sinsy用)を読み取る

    source: path, lines
    """
    full_label = HTSFullLabel()
    return full_label.load(source)


def notenum_to_abspitch(notenum) -> str:
    """
    音高をC4のような記法に変換する
    """
    return NOTENUM_TO_ABSPITCH_DICT[str(notenum)]


def abspitch_to_notenum(abspitch: str):
    """
    音高をノート番号に変換する
    """
    return ABSPITCH_TO_NOTENUM_DICT[abspitch]


class HTSFullLabel(UserList):
    """
    HTSのフルコンテキストラベルの1行を扱うクラス
    OneLine からなる list
    [OneLine, OneLine, ..., OneLine]
    """

    def __init__(self, init=None):
        super().__init__(init)
        self.song = Song()

    def as_mono(self) -> _label.Label:
        """
        モノラベルに変換する
        """
        # TODO: 内包表記かlambda式で高速化
        mono_label = _label.Label()
        for oneline in self:
            mono_phoneme = _label.Phoneme()
            mono_phoneme.start = copy(oneline.phoneme.start)
            mono_phoneme.end = copy(oneline.phoneme.end)
            mono_phoneme.symbol = copy(oneline.phoneme.identity)
            mono_label.append(mono_phoneme)
        return mono_label

    def write(self, path, strict_sinsy_style: bool = True, as_mono: bool = False,
              mode='w', encoding='utf-8') -> str:
        """
        ファイル出力する
        strict_sinsy_style: bool:
            「休符の長さ」が前後の発声に影響するかどうかを左右する。
            Trueのときは d, f における休符の長さ情報が削除されて 'xx' になる。
            Falseのときは d, f における休符の長さ情報が維持される。
        """
        # 休符周辺の仕様をSinsyに近づける。
        new_label = adjust_pau_contexts(self, strict=strict_sinsy_style)
        # 文字列にする
        s = '\n'.join(list(map(str, new_label)))

        # ファイル出力
        with open(path, mode=mode, encoding=encoding, newline='\n') as f:
            f.write(s)
        return s

    def load(self, source, encoding='utf-8'):
        """
        ファイル、文字列、文字列のリスト、Songオブジェクトのいずれかより値を取得して登録する。
        """
        if isinstance(source, str):
            self._load_from_path(source, encoding=encoding)
            self.generate_songobj()
        elif isinstance(source, Song):
            self._load_from_songobj(source)
            self.fill_contexts_from_songobj()
        elif isinstance(source, list):
            self._load_from_lines(source)
            self.generate_songobj()
        else:
            raise TypeError(f'Type of the argument "source" must be str, list or {Song}.')
        return self

    def _load_from_path(self, path, encoding: str = 'utf-8'):
        """
        ファイルをもとに値を登録する。
        """
        # パスに半角スペースが入っている場合に出現する引用符を除去
        path = path.strip('"')
        # ファイルを読み取って行のリストにする
        try:
            with open(path, mode='r', encoding=encoding) as f:
                lines = [line.rstrip('\r\n') for line in f.readlines()]
        except UnicodeDecodeError:
            with open(path, mode='r', encoding='cp932') as f:
                lines = [line.rstrip('\r\n') for line in f.readlines()]
        # 行ごとに分割したリストをもとに情報を登録する。
        self._load_from_lines(lines)
        return self

    def _load_from_lines(self, lines: list):
        """
        文字列のリスト(行のリスト)をもとに値を登録する。
        """
        # 各行を解析してHTSFullLabelに追加する。
        for line in lines:
            # 1行分の情報用のオブジェクトを生成
            ol = OneLine()
            # 正規表現で上手く区切れない文字を置換する
            # 空白で分割して、時刻情報とそれ以外のコンテキストに分ける
            line_split = line.split(maxsplit=2)
            ol.start = int(line_split[0])
            ol.end = int(line_split[1])
            str_contexts = line_split[2]
            # コンテキスト文字列を /A: などの文字列で区切って一次元リストにする
            l_contexts = re.split('/.:', str_contexts)
            # 特定の文字でさらに区切って二次元リストにする
            sep = re.escape('=+-~∼!@#$%^ˆ&;_|[]')
            l_contexts_2d = [re.split((f'[{sep}]'), s) for s in l_contexts]
            # 1行分の情報用のオブジェクトに、各種コンテキストを登録する
            ol.p, ol.a, ol.b, ol.c, ol.d, ol.e, ol.f, ol.g, ol.h, ol.i, ol.j = l_contexts_2d
            # 1行分の情報用のオブジェクトを HTSFullLabel オブジェクトに追加する。
            self.append(ol)
        return self

    def _load_from_songobj(self, songobj: list):
        """
        hts.Song オブジェクトをもとにコンテキストを登録する。
        """
        self.song = songobj
        return self

    def fill_contexts_from_songobj(self):
        """
        自身が持つSongオブジェクトをもとにコンテキスト情報を埋める。
        """
        song = self.song
        dummy_note = Note()
        dummy_syllable = Syllable()
        dummy_syllable.append(Phoneme())
        dummy_note.append(dummy_syllable)

        notes = [dummy_note] + song + [deepcopy(dummy_note)]

        onelines = []
        for i_n, note in enumerate(notes[1:-1], 1):
            syllables = notes[i_n - 1] + note + notes[i_n + 1]
            for i_s, syllable in enumerate(syllables[1:-1], 1):
                for phoneme in syllable:
                    ol = OneLine()
                    # ノート情報を登録
                    ol.previous_note = notes[i_n - 1]
                    ol.note = note
                    ol.next_note = notes[i_n + 1]
                    # 音節情報を登録
                    ol.previous_syllable = syllables[i_s - 1]
                    ol.syllable = syllable
                    ol.next_syllable = syllables[i_s + 1]
                    # 音素情報を登録(現在の音素のみ)
                    ol.phoneme = phoneme
                    # 楽曲情報を登録
                    ol.song = song
                    # print(list(map(id, [song, note, syllable, phoneme])))
                    onelines.append(ol)
        self.data = onelines
        self.fill_phonemes()
        return self

    def fill_phonemes(self):
        """
        Phoneme をもとに、前後の音素に関する項を埋める。
        """
        extended_self = [OneLine(), OneLine()] + self.data + [OneLine(), OneLine()]
        # ol is OneLine objec
        for i, ol in enumerate(extended_self[2:-2], 2):
            ol.before_previous_phoneme = extended_self[i - 2].phoneme
            ol.previous_phoneme = extended_self[i - 1].phoneme
            ol.next_phoneme = extended_self[i + 1].phoneme
            ol.after_next_phoneme = extended_self[i + 2].phoneme
        return self

    def generate_songobj(self):
        """
        フルコンテキストラベルの情報をもとにSongオブジェクトをつくる。
        self.song に格納する。
        """
        # Song は1つの HTSFullLabel につき1つだけなので、ループする必要がない。
        # Song は一番最初の行の情報がすべての行に通用すると考える。
        song = self[0].song
        song.data = []
        note = Note()
        syllable = Syllable()

        for ol in self:
            phoneme = ol.phoneme
            # すでに入っている音節や音素を取り除く。取り除かないと数が倍になる。
            ol.note.data = []
            ol.syllable.data = []
            # 休符のときは無条件でノートと音節を切り替える。
            if ol.phoneme.is_rest():
                note = ol.note
                song.append(note)
                syllable = ol.syllable
                note.append(syllable)
            # 休符以外のとき
            else:
                # 処理する行が「ノート内で1番最初の音節」かつ
                # 「音節内でいちばん最初の音素」のとき、ノートを切り替える。
                if str(ol.syllable.position) == '1' and str(phoneme.position) == '1':
                    note = ol.note
                    song.append(note)
                # 処理する行が「音節内で1番最初の音素」なとき、音節を切り替える。
                if str(phoneme.position) == '1':
                    syllable = ol.syllable
                    note.append(syllable)
            syllable.append(phoneme)

        song.autofill()
        self.song = song
        self.fill_contexts_from_songobj()


class OneLine:
    """
    HTSのフルコンテキストラベルの1行を扱うクラス
    ファイルを読み取って HTSFullLabel を生成するときと、
    HTSFullLabel をファイル入出力するときに使う。
    """

    def __init__(self):
        self.before_previous_phoneme = Phoneme()
        self.previous_phoneme = Phoneme()
        self.phoneme = Phoneme()
        self.next_phoneme = Phoneme()
        self.after_next_phoneme = Phoneme()
        self.previous_syllable = Syllable()
        self.syllable = Syllable()
        self.next_syllable = Syllable()
        self.previous_note = Note()
        self.note = Note()
        self.next_note = Note()
        self.previous_phrase = Phrase()
        self.phrase = Phrase()
        self.next_phrase = Phrase()
        self.song = Song()

    def __str__(self):
        str_self = ''.join((
            f'{self.start} {self.end} ',
            # Phoneme 関連
            '{}@{}^{}-{}+{}={}_{}%{}^{}_{}~{}-{}!{}[{}${}]{}'
            .format(*self.p),
            # Syllable 関連
            '/A:{}-{}-{}@{}~{}'.format(*self.a),
            '/B:{}_{}_{}@{}|{}'.format(*self.b),
            '/C:{}+{}+{}@{}&{}'.format(*self.c),
            # Note 関連
            '/D:{}!{}#{}${}%{}|{}&{};{}-{}'.format(*self.d),
            '/E:{}]{}^{}={}~{}!{}@{}#{}+{}]{}${}|{}[{}&{}]{}={}^{}~{}#{}_{};{}${}&{}%{}[{}|{}]{}-{}^{}+{}~{}={}@{}${}!{}%{}#{}|{}|{}-{}&{}&{}+{}[{};{}]{};{}~{}~{}^{}^{}@{}[{}#{}={}!{}~{}+{}!{}^{}' \
            .format(*self.e),
            '/F:{}#{}#{}-{}${}${}+{}%{};{}'.format(*self.f),
            # Phrase 関連
            '/G:{}_{}'.format(*self.g),
            '/H:{}_{}'.format(*self.h),
            '/I:{}_{}'.format(*self.i),
            # Song 関連
            '/J:{}~{}@{}'.format(*self.j)
        ))
        return str_self

    @property
    def start(self) -> int:
        """
        発声開始時刻
        """
        return int(self.phoneme.start)

    @start.setter
    def start(self, start_time: int):
        self.phoneme.start = start_time

    @property
    def end(self) -> int:
        """
        発声終了時刻
        """
        return int(self.phoneme.end)

    @end.setter
    def end(self, end_time: int):
        self.phoneme.end = end_time

    @property
    def duration(self) -> int:
        """
        発声時間の長さ (発声終了時刻-発声開始時刻)
        """
        return self.phoneme.end - self.phoneme.start

    @property
    def symbol(self) -> str:
        """
        音素記号
        """
        return self.phoneme.identity

    @property
    def p(self) -> list:
        """
        音素に関するコンテキスト
        """
        p = [
            self.phoneme.language_independent_identity,
            self.before_previous_phoneme.identity,
            self.previous_phoneme.identity,
            self.phoneme.identity,
            self.next_phoneme.identity,
            self.after_next_phoneme.identity,
            self.before_previous_phoneme.flag,
            self.previous_phoneme.flag,
            self.phoneme.flag,
            self.next_phoneme.flag,
            self.after_next_phoneme.flag,
            self.phoneme.position,
            self.phoneme.position_backward,
            self.phoneme.distance_from_previous_vowel,
            self.phoneme.distance_to_next_vowel,
            self.phoneme.undefined_context
        ]
        return p

    @p.setter
    def p(self, phoneme_contexts: list):
        (self.phoneme.language_independent_identity,
         self.before_previous_phoneme.identity,
         self.previous_phoneme.identity,
         self.phoneme.identity,
         self.next_phoneme.identity,
         self.after_next_phoneme.identity,
         self.before_previous_phoneme.flag,
         self.previous_phoneme.flag,
         self.phoneme.flag,
         self.next_phoneme.flag,
         self.after_next_phoneme.flag,
         self.phoneme.position,
         self.phoneme.position_backward,
         self.phoneme.distance_from_previous_vowel,
         self.phoneme.distance_to_next_vowel,
         self.phoneme.undefined_context
         ) = phoneme_contexts[0:16]

    @property
    def a(self) -> list:
        """
        直前の音節のコンテキスト
        """
        return self.previous_syllable.contexts

    @a.setter
    def a(self, syllable_contexts: list):
        self.previous_syllable.contexts = syllable_contexts

    @property
    def b(self) -> list:
        """
        現在の音節のコンテキスト
        """
        return self.syllable.contexts

    @b.setter
    def b(self, syllable_contexts: list):
        self.syllable.contexts = syllable_contexts

    @property
    def c(self) -> list:
        """
        直後の音節のコンテキスト
        """
        return self.next_syllable.contexts

    @c.setter
    def c(self, syllable_contexts: list):
        self.next_syllable.contexts = syllable_contexts

    @property
    def d(self) -> list:
        """
        直前のノートのコンテキスト
        """
        return self.previous_note.contexts

    @d.setter
    def d(self, note_contexts: list):
        self.previous_note.contexts = note_contexts

    @property
    def e(self) -> list:
        """
        現在のノートのコンテキスト
        """
        return self.note.contexts

    @e.setter
    def e(self, note_contexts: list):
        self.note.contexts = note_contexts

    @property
    def f(self) -> list:
        """
        直後のノートのコンテキスト
        """
        return self.next_note.contexts

    @f.setter
    def f(self, note_contexts: list):
        self.next_note.contexts = note_contexts

    @property
    def g(self) -> list:
        """
        直前のフレーズのコンテキスト
        """
        return self.previous_phrase.contexts

    @g.setter
    def g(self, phrase_contexts: list):
        self.previous_phrase.contexts = phrase_contexts

    @property
    def h(self) -> list:
        """
        現在のフレーズのコンテキスト
        """
        return self.phrase.contexts

    @h.setter
    def h(self, phrase_contexts: list):
        self.phrase.contexts = phrase_contexts

    @property
    def i(self) -> list:
        """
        直後のフレーズのコンテキスト
        """
        return self.next_phrase.contexts

    @i.setter
    def i(self, phrase_contexts: list):
        self.next_phrase.contexts = phrase_contexts

    @property
    def j(self) -> list:
        """
        (現在の)楽曲のコンテキスト
        1ファイル1曲。
        """
        return self.song.contexts

    @j.setter
    def j(self, song_contexts: list):
        self.song.contexts = song_contexts


class Song(UserList):
    """
    曲を扱うクラス
    今日の曲(j1-j3)
    [Phrase, Phrase, ..., Phrase]
    [Note, Note, ..., Note]
    """

    def __init__(self, init=None):
        super().__init__(init)
        self.contexts = ['xx'] * 3
        self.number_of_measures = 'xx'

    # @property
    # def all_phrases(self):
    #     """
    #     Phraseをすべて並べたリストを返す。
    #     """
    #     return self

    @property
    def all_notes(self):
        """
        Noteをすべて並べたリストを返す。
        """
        return self.data

    @property
    def all_syllables(self):
        """
        Syllableをすべて並べたリストを返す。
        """
        return list(chain.from_iterable(self.all_notes))

    @property
    def all_phonemes(self):
        """
        Phonemeをすべて並べたリストを返す。
        """
        return list(chain.from_iterable(self.all_syllables))

    @property
    def number_of_phrases(self):
        """
        楽曲内のフレーズ数。(j3)
        """
        return self.contexts[2]

    @number_of_phrases.setter
    def number_of_phrases(self, number: int):
        self.contexts[2] = number

    def as_mono(self) -> _label.Label:
        """
        モノラベルに変換する
        """
        mono_label = _label.Label()
        for full_phoneme in self.all_phonemes:
            mono_phoneme = _label.Phoneme()
            mono_phoneme.start = copy(full_phoneme.start)
            mono_phoneme.end = copy(full_phoneme.end)
            mono_phoneme.symbol = copy(full_phoneme.identity)
            mono_label.append(mono_phoneme)
        return mono_label

    def write(self, path, strict_sinsy_style: bool = True, as_mono: bool = False,
              mode='w', encoding='utf-8') -> Union[HTSFullLabel, _label.Label]:
        """
        ファイル出力する。
        HTSFullLabelからではなく、USTなどからSongが直接生成されている場合に対応する。
        """
        full_label = HTSFullLabel()
        full_label.song = self

        full_label.fill_contexts_from_songobj()
        if as_mono is True:
            mono_label = full_label.as_mono()
            mono_label.write(path, mode=mode, encoding=encoding)
            return mono_label

        full_label.write(
            path, strict_sinsy_style=strict_sinsy_style,
            mode=mode, encoding=encoding)
        return full_label

    def reset_time(self):
        """
        ノート長をもとに、Sinsyの出力と同じになるように、
        Phonemeオブジェクトのstartとendを計算して登録する。

        単位は100ns
        100ns単位での長さがすでに計算されているものとする。(note.length_100ns)
        """
        t_start = 0
        t_end = 0
        for note in self.all_notes:
            phonemes_in_note = tuple(chain.from_iterable(note))
            t_end += note.length_100ns
            for phoneme in phonemes_in_note:
                phoneme.start = Decimal(t_start).quantize(Decimal('0'), rounding=ROUND_HALF_UP)
                phoneme.end = Decimal(t_end).quantize(Decimal('0'), rounding=ROUND_HALF_UP)
            t_start = t_end

    def autofill(self, hts_conf: Union[None, dict] = None):
        """
        自動補完可能なものをすべて自動補完する。
        """
        # この順でやらないと、p1が未設定の状態で Note.is_rest() をやってしまう
        self._fill_phoneme_contexts(hts_conf)
        self._fill_syllable_contexts()
        self._fill_note_contexts()
        self._fill_song_contexts()

    def _fill_phoneme_contexts(self, hts_conf: Union[None, dict]):
        """
        p1, p12, p13, p14, p15 を補完する。
        """
        # p12, p13, p14, p15を埋める
        if hts_conf is None:
            self._fill_p1(vowels=VOWELS, pauses=PAUSES, silences=SILENCES, breaks=BREAKS)
        else:
            self._fill_p1(vowels=hts_conf['VOWELS'], pauses=hts_conf['PAUSES'],
                          silences=hts_conf['SILENCES'], breaks=hts_conf['BREAKS'])
        self._fill_p12_p13()
        self._fill_p14_p15()

    def _fill_p1(self, vowels, pauses, silences, breaks):
        # p1 を埋める
        for phoneme in self.all_phonemes:
            phoneme_identity = phoneme.identity
            if phoneme_identity == 'xx':
                phoneme.language_independent_identity = 'xx'
            elif phoneme_identity in vowels:
                phoneme.language_independent_identity = 'v'
            elif phoneme_identity in pauses:
                phoneme.language_independent_identity = 'p'
            elif phoneme_identity in silences:
                phoneme.language_independent_identity = 's'
            elif phoneme_identity in breaks:
                phoneme.language_independent_identity = 'b'
            else:
                phoneme.language_independent_identity = 'c'

    def _fill_p12_p13(self):
        """
        p12 と p13 (音節内での位置) を補完する。
        """
        for syllable in self.all_syllables:
            len_syllable = len(syllable)
            for i, phoneme in enumerate(syllable):
                # p12
                phoneme.position = i + 1
                # p13
                phoneme.position_backward = len_syllable - i

    def _fill_p14_p15(self):
        """
        p14 と p15 (母音からの距離) を補完する。
        """
        # p14
        for syllable in self.all_syllables:
            distance = None
            for phoneme in syllable:
                if phoneme.is_vowel():
                    phoneme.distance_from_previous_vowel = 'xx'
                    distance = 1
                elif distance is None:
                    continue
                elif phoneme.is_consonant():
                    phoneme.distance_from_previous_vowel = distance
                    distance += 1
                # 母音でも子音でもない場合(clとか)
                else:
                    distance += 1
        # p15
        for syllable in self.all_syllables:
            distance = None
            for phoneme in reversed(syllable):
                if phoneme.is_vowel():
                    phoneme.distance_to_next_vowel = 'xx'
                    distance = 1
                elif distance is None:
                    continue
                elif phoneme.is_consonant():
                    phoneme.distance_to_next_vowel = distance
                    distance += 1
                # 母音でも子音でもない場合(clとか)
                else:
                    distance += 1

    def _fill_syllable_contexts(self):
        """
        b1, b2, b3 を補完する。
        """
        for note in self.all_notes:
            len_note = len(note)
            for i, syllable in enumerate(note):
                # b1
                syllable.number_of_phonemes = len(syllable)
                # b2
                syllable.position = i + 1
                # b3
                syllable.position_backward = len_note - i

    def _fill_note_contexts(self):
        """e を補完する。

        必要なデータ:
            - e1: 絶対音高
            - e5: テンポ
            - e8: ノート長(96分音符)
        補完するデータ:
            - e6: ノート内音節数
            - e7: ノート長(10ms)
            - e18-e25: フレーズ内での位置
                - e20, e21: フレーズ内での位置(100ms)
                - e22, e23: フレーズ内での位置(96分音符)
                - e24, e25: フレーズ内での位置(パーセント)
            - Note.position_100ns
            - Note.position_100ns_backward
            - e57-e58: 前後のノートとの音高差
        """
        # e6, e7 を埋める。
        for note in self.all_notes:
            # e6 (ノート内音節数) を埋める
            note.number_of_syllables = len(note)
            # e8 の情報をもとに e7 を埋める。テンポ情報(e5)がないと困る。
            # 100ns単位での長さも登録する。
            if note.tempo != 'xx' and note.length != 'xx':
                note.length_10ms = Decimal(note.length_100ns / 100000
                                           ).quantize(Decimal('0'), rounding=ROUND_HALF_UP)

        # フレーズ内で何番目のノートかを埋める
        self._fill_e18_e19()
        # フレーズ内での 96分音符単位での位置を埋める。e18,e19が必要。
        self._fill_e22_e23()
        # フレーズ内での 100ms単位での位置を埋める。e18,e19が必要。
        # フレーズ内での 100ns単位での位置も埋める(utaupy独自コンテキスト)。
        self._fill_e20_e21_100ns()
        # フレーズ内での パーセント表記での位置を埋める。e18, e19, 100ns位置が必要
        self._fill_e24_e25()
        # 音高の前後変化を埋める
        self._fill_e57_e58()

    def _fill_e18_e19(self):
        """
        「Phrase内で何番目か(e18, e19)」の項を埋める。
        Phraseの扱いは難しいので、「休符からの距離」を代わりに用いる。
        """
        counter = 'xx'
        notes = self.all_notes

        # 最初のノートが休符でなくても動作するようにする。
        counter = 0
        # 直前の休符からの距離を数える
        for note in notes:
            # 休符のときは距離をリセット
            if note.is_rest():
                counter = 0
                note.position = 'xx'
            # 休符ではなく、休符位置が分かっているときは登録
            else:
                counter += 1
                note.position = counter

        # 最後のノートが休符でなくても動作するようにする。
        counter = 0
        # 次の休符までの距離を登録する。
        for note in reversed(notes):
            if note.is_rest():
                counter = 0
                note.position_backward = 'xx'
            else:
                counter += 1
                note.position_backward = counter

    def _fill_e20_e21_100ns(self):
        """
        「フレーズ内での位置(100ms単位での表記) (e20, e21)」の項を埋める。

        e18とe19のデータがある前提で実行する。
        10msでの累計だとずれるので100nsで計算してから10msに直して登録する。
        """

        counter_100ns = 0

        # 前の休符から数えた時間を登録する。
        for note in self.all_notes:
            # 休符のときは 'xx'
            if note.position == 'xx':
                note.position_100ms = 'xx'
                note.position_100ns = None
                counter_100ns = 0
            # フレーズ中で最初のノートのときは時間をリセットして、累積時間を増やす。
            elif note.position == 1:
                note.position_100ms = 0
                note.position_100ns = 0
                counter_100ns += note.length_100ns
            # 休符でもなくて最初のノートでもないとき
            else:
                note.position_100ms = Decimal(
                    counter_100ns / 1000000).quantize(Decimal('0'), rounding=ROUND_HALF_UP)
                note.position_100ns = counter_100ns
                counter_100ns += note.length_100ns

        counter_100ns = 0
        # 次の休符から数えた時間を登録する。
        for note in reversed(self.all_notes):
            # 休符のときは 'xx'
            if note.position_backward == 'xx':
                counter_100ns = 0
                note.position_100ms_backward = 'xx'
                note.position_100ns_backward = None
            # フレーズ中で最後のノートのときは時間をリセットして、累積時間を増やす。
            elif note.position_backward == 1:
                counter_100ns = note.length_100ns
                note.position_100ms_backward = Decimal(
                    counter_100ns / 1000000).quantize(Decimal('0'), rounding=ROUND_HALF_UP)
                note.position_100ns_backward = counter_100ns
            # 休符でもなくて最後のノートでもないとき
            else:
                counter_100ns += note.length_100ns
                note.position_100ms_backward = Decimal(
                    counter_100ns / 1000000).quantize(Decimal('0'), rounding=ROUND_HALF_UP)
                note.position_100ns_backward = counter_100ns

    def _fill_e22_e23(self):
        """「フレーズ内での位置(96分音符単位) (e22, e23)」の項を埋める。

        e18とe19のデータがある前提で実行する。
        """
        counter = 0
        # 前の休符から数えた時間を登録する。
        for note in self.all_notes:
            # 休符のときは 'xx'
            if note.position == 'xx':
                note.contexts[21] = 'xx'
                counter = 0
            # フレーズ中で最初のノートのときは時間をリセットして、累積時間を増やす。
            elif note.position == 1:
                note.contexts[21] = 0
                counter += int(note.length)
            # 休符でもなくて最初のノートでもないとき
            else:
                note.contexts[21] = counter
                counter += int(note.length)

        counter = 0
        # 次の休符から数えた時間を登録する。
        for note in reversed(self.all_notes):
            # 休符のときは 'xx'
            if note.position_backward == 'xx':
                counter = 0
                note.contexts[22] = 'xx'
            # フレーズ中で最後のノートのときは時間をリセットして、累積時間を増やす。
            elif note.position_backward == 1:
                counter = int(note.length)
                note.contexts[22] = counter
            # 休符でもなくて最後のノートでもないときは普通に登録して、累積時間を増やす。
            else:
                counter += int(note.length)
                note.contexts[22] = counter

    def _fill_e24_e25(self):
        """
        「フレーズ内での位置(パーセント表記) (e24, e25)」の項を埋める。
        e18, e19, length_100ns のデータがある前提で実行する。
        """
        notes = self.all_notes
        # フレーズの全体の長さ
        phrase_length_100ns = 0
        counter_100ns = 0
        for note in notes:
            if note.position == 'xx' or note.is_rest():
                phrase_length_100ns = 0
                counter_100ns = 0
                note.contexts[23] = 'xx'
            elif note.position == 1:
                phrase_length_100ns = note.position_100ns_backward
                note.contexts[23] = Decimal(100 * counter_100ns / phrase_length_100ns
                                            ).quantize(Decimal('0'), rounding=ROUND_HALF_UP)
                counter_100ns += note.length_100ns
            else:
                note.contexts[23] = Decimal(100 * counter_100ns / phrase_length_100ns
                                            ).quantize(Decimal('0'), rounding=ROUND_HALF_UP)
                counter_100ns += note.length_100ns

        # NOTE: 次のブロックの処理でうまくいかなかった場合、コメント部分のコードを使う。
        for note in notes:
            position_percent = note.contexts[23]
            if position_percent == 'xx':
                note.contexts[24] = 'xx'
            else:
                note.contexts[24] = 100 - note.contexts[23]
        # NOTE: 上のブロックが動かなかったらここを使う
        # # フレーズの全体の長さ
        # phrase_length_100ns = 0
        # counter_100ns = 0
        # for note in reversed(self.all_notes):
        #     # 休符のときは 'xx'
        #     if note.position_backward == 'xx' or note.is_rest():
        #         phrase_length_100ns = 0
        #         counter_100ns = 0
        #         note.contexts[24] = 'xx'
        #     # フレーズ中で最後のノートのときは時間をリセットして、累積時間を増やす。
        #     elif note.position_backward == 1:
        #         phrase_length_100ns = note.position_100ns + note.length_100ns
        #         counter_100ns = note.length_100ns
        #         note.contexts[24] = Decimal(100 * counter_100ns / phrase_length_100ns
        #                                     ).quantize(Decimal('0'), rounding=ROUND_HALF_UP)
        #     # 休符でもなくて最後のノートでもないときは普通に登録して、累積時間を増やす。
        #     else:
        #         counter_100ns += note.length_100ns
        #         note.contexts[24] = Decimal(100 * counter_100ns / phrase_length_100ns
        #                                     ).quantize(Decimal('0'), rounding=ROUND_HALF_UP)

    def _fill_e57_e58(self):
        """
        「前後のノートとの音高差 (e57, e58)」の項を埋める。
        e1がある前提で実行する。
        休符が入ると死ぬw
        """
        notes = self.all_notes
        # e57を埋める
        for i, note in enumerate(notes[1:], 1):
            previous_note = notes[i - 1]
            previous_abspitch = previous_note.absolute_pitch
            current_abspitch = note.absolute_pitch
            # 直前のノートまたは今のノートが休符のときはスキップ
            if (previous_note.is_rest() or note.is_rest()):
                note.contexts[56] = 'xx'
                continue
            # 直前のノートも今のノートも音符のとき
            previous_notenum = abspitch_to_notenum(previous_abspitch)
            current_notenum = abspitch_to_notenum(current_abspitch)
            pitch_difference = previous_notenum - current_notenum
            note.contexts[56] = f'{"p" if pitch_difference >= 0 else "m"}{abs(pitch_difference)}'

        # e58 を埋める
        # 処理内容を比較しやすいようにprevious_noteにしているが、実際はnext_note
        for i, note in enumerate(reversed(notes[:-1]), 1):
            previous_note = notes[-i]
            previous_abspitch = previous_note.absolute_pitch
            current_abspitch = note.absolute_pitch
            # 直後のノートまたは今のノートが休符のときはスキップ
            if (previous_note.is_rest() or note.is_rest()):
                note.contexts[57] = 'xx'
                continue
            # 直前のノートも今のノートも音符のとき
            previous_notenum = abspitch_to_notenum(previous_abspitch)
            current_notenum = abspitch_to_notenum(current_abspitch)
            pitch_difference = previous_notenum - current_notenum
            note.contexts[57] = f'{"p" if pitch_difference >= 0 else "m"}{abs(pitch_difference)}'

    def _fill_song_contexts(self):
        """
        Songオブジェクトのコンテキストを自動補完する。
        """
        self._fill_j3()

    def _fill_j3(self):
        """
        Songオブジェクト内のフレーズ数(j3) の項を埋める。
        """
        notes = self.all_notes
        previous_note_is_rest = notes[0].is_rest()

        counter = 0
        # 最初が音符だった時はフレーズ数1からスタート
        if not previous_note_is_rest:
            counter = +1
        # 休符→音符 の並びの回数を検出する。
        for note in notes[1:]:
            current_note_is_rest = note.is_rest()
            if previous_note_is_rest and not current_note_is_rest:
                counter += 1
            previous_note_is_rest = current_note_is_rest
        self.number_of_phrases = counter


class Phrase(UserList):
    """フレーズを扱うクラス

    昨日のフレーズ G (g1~g2)
    今日のフレーズ H (h1~h2)
    明日のフレーズ I (i1~i2)
    基本的にはHのみを操作する。
    [Note, Note, Note, ..., Note]
    """

    def __init__(self, init=None):
        super().__init__(init)
        self.contexts = ['xx'] * 2

    @property
    def number_of_syllables(self):
        """
        フレーズ内の音節数。

        OneLine からSongを生成するときに使うかも。
        フレーズの切り替わり判定をするときに使うかも。
        多分使わない。
        """
        return self.contexts[0]

    @number_of_syllables.setter
    def number_of_syllables(self, number):
        self.contexts[0] = number

    @property
    def number_of_phonemes(self):
        """
        フレーズ内の音素数。
        OneLine からSongを生成するときに使う。
        """
        return self.contexts[1]

    @number_of_phonemes.setter
    def number_of_phonemes(self, number):
        self.contexts[1] = number


class Note(UserList):
    """
    ノートまたは休符を扱うクラス
    1ノート（ノートと休符）を扱うクラス
    昨日のノート D (d1~d5)
    今日のノート E (e1~e60)
    明日のノート F (f1~f5)
    基本的にはEのみを操作する。
    [Syllable, Syllable, ..., Syllable]
    """

    def __init__(self, init=None):
        super().__init__(init)
        self.contexts = ['xx'] * 60
        self.position_100ns: int = None
        self.position_100ns_backward: int = None

    @property
    def absolute_pitch(self):
        """
        ノートの絶対音高 (C0-G9) (e1)
        # TODO: 番号でもsetできるようにする。
        """
        return self.contexts[0]

    @absolute_pitch.setter
    def absolute_pitch(
            self, absolute_pitch: str = None, notenum: int = None):
        """
        absolute_pitch: C0-G9 のような表記
        notenum: MIDIなどの note number による表記(C-1:0 ~ G9:127)
        """
        # どちらか片方は必須だし、
        # 片方しか利用できませんよというやつ
        assert not (absolute_pitch is None) == (notenum is None)
        if absolute_pitch is not None:
            self.contexts[0] = absolute_pitch
        else:
            self.contexts[0] = notenum_to_abspitch(notenum)

    @property
    def relative_pitch(self) -> int:
        """
        ノートの相対音高(p2)
        """
        return self.contexts[1]

    @relative_pitch.setter
    def relative_pitch(self, relative_pitch: int):
        self.contexts[1] = relative_pitch

    @property
    def key(self):
        """
        ノートのキー(p3)
        """
        return self.contexts[2]

    @key.setter
    def key(self, key: int):
        self.contexts[2] = key

    @property
    def beat(self):
        """
        拍子情報(p4)
        3/4 とか 4/4 とか 6/8 とか
        """
        return self.contexts[3]

    @beat.setter
    def beat(self, beat: str):
        self.contexts[3] = str(beat)

    @property
    def notenum(self):
        """
        音高をノート番号で取得する
        ラベルファイルには記録されないが、
        C4のような表記でe1に記録される。
        """
        return abspitch_to_notenum(self.absolute_pitch)

    @notenum.setter
    def notenum(self, notenum):
        self.absolute_pitch = notenum_to_abspitch(notenum)

    @property
    def tempo(self):
        """
        テンポ(e5)
        """
        return self.contexts[4]

    @tempo.setter
    def tempo(self, tempo: int):
        self.contexts[4] = tempo

    @property
    def number_of_syllables(self):
        """
        ノート内音節数(e6)
        """
        return self.contexts[5]

    @number_of_syllables.setter
    def number_of_syllables(self, number: int):
        self.contexts[5] = number

    @property
    def length_10ms(self):
        """
        ノート長(e7)
        10ms単位の整数で計算
        """
        return self.contexts[6]

    @length_10ms.setter
    def length_10ms(self, length: int):
        self.contexts[6] = length

    @property
    def length(self):
        """
        ノート長(e8)
        96分音符いくつ分かで計算
        """
        return self.contexts[7]

    @length.setter
    def length(self, length: int):
        self.contexts[7] = length

    @property
    def length_100ns(self):
        """
        ノート長(ラベル出力なし)
        100ns単位の整数で計算
        """
        if 'xx' in (self.tempo, self.length):
            return 'xx'
        return float(25000000 * int(self.length) / int(self.tempo))

    @property
    def position(self):
        """
        フレーズ内での位置(e18)
        """
        return self.contexts[17]

    @position.setter
    def position(self, position: int):
        self.contexts[17] = position

    @property
    def position_backward(self):
        """
        フレーズ内での後ろから数えた位置(e19)
        1-indexed
        """
        return self.contexts[18]

    @position_backward.setter
    def position_backward(self, position_backward: int):
        self.contexts[18] = position_backward

    @property
    def position_100ms(self):
        """
        フレーズ内での位置(e20)
        100ms単位
        """
        return self.contexts[19]

    @position_100ms.setter
    def position_100ms(self, position: int):
        self.contexts[19] = position

    @property
    def position_100ms_backward(self):
        """
        フレーズ内での後ろから数えた位置(e21)
        100ms単位
        """
        return self.contexts[20]

    @position_100ms_backward.setter
    def position_100ms_backward(self, position_backward: int):
        self.contexts[20] = position_backward

    @property
    def phonemes(self) -> list:
        """
        Note内のすべてのPhonemeオブジェクトを並べたリストを返す。
        """
        return list(chain.from_iterable(self))

    def is_rest(self):
        """
        休符かどうかを返す
        ノート内の最初の音素が休符かどうかで判断する
        """
        return self[0][0].is_rest()

    def is_pau(self):
        """
        後方互換のために残している。
        'is_pau()' is now 'is_rest()'
        """
        return self.is_rest()


class Syllable(UserList):
    """
    1音節を扱うクラス
    昨日の音節 A (a1~a5)
    今日の音節 B (b1~b5)
    明日の音節 C (c1~c5)
    [Phoneme, Phoneme, ..., Phoneme]
    """

    def __init__(self, init=None):
        super().__init__(init)
        self.contexts = ['xx'] * 5

    @property
    def number_of_phonemes(self):
        """
        音節内の音素数(b1)
        """
        return self.contexts[0]

    @number_of_phonemes.setter
    def number_of_phonemes(self, number: int):
        self.contexts[0] = number

    @property
    def position(self):
        """
        ノート内での位置(b2)
        1-indexed
        """
        return self.contexts[1]

    @position.setter
    def position(self, position: int):
        self.contexts[1] = position

    @property
    def position_backward(self):
        """

        ノート内での後ろから数えた位置(b3)
        1-indexed
        """
        return self.contexts[2]

    @position_backward.setter
    def position_backward(self, position_backward: int):
        self.contexts[2] = position_backward


class Phoneme:
    """
    音素を扱うクラス
    p1~p16
    """

    def __init__(self):
        # 発声開始時刻
        self.start: int = 0
        # 発声終了時刻
        self.end: int = 0
        # p1: 言語非依存の音素記号(p, c, v など)
        self.language_independent_identity: str = 'xx'
        # p4: 音素記号
        self.identity: str = 'xx'
        # p9: フラグ
        self.flag: str = 'xx'
        # p12: 音節(Syllable)内の、先頭から数えた位置
        self.position: int = 'xx'
        # p13: 音節(Syllable)内の、末尾から数えた位置
        self.position_backward: int = 'xx'
        # p14: 今が子音のときのみに定義される、直前の母音からの距離
        self.distance_from_previous_vowel: int = 'xx'
        # p15: 今が子音のときのみに定義される、直後の母音までの距離
        self.distance_to_next_vowel: int = 'xx'
        # 未定義のコンテキスト
        self.undefined_context = 'xx'

    def __str__(self):
        return f'{self.start} {self.end} {self.identity}'

    @property
    def duration(self) -> int:
        """
        discription
        """
        return self.end - self.start

    def is_vowel(self):
        """
        母音かどうか
        """
        return self.language_independent_identity == 'v'

    def is_consonant(self):
        """
        子音かどうか
        """
        return self.language_independent_identity == 'c'

    def is_rest(self):
        """
        休符かどうか
        """
        return self.language_independent_identity in ('s', 'p')

    def is_pau(self):
        """休符かどうかを判定

        後方互換のために残した。
        'is_pau' is now 'is_rest'
        """
        return self.is_rest()

    def is_break(self):
        """
        息継ぎかどうか
        """
        return self.language_independent_identity == 'b'


def adjust_pau_contexts(full_label: HTSFullLabel, strict: bool = True) -> HTSFullLabel:
    """
    出力用に休符まわりの音節コンテキストを調整する。
    """
    # deepcopyが深く、できるだけSinsyの出力に近づける。でも遅い。
    if strict:
        new_label = HTSFullLabel()
        # NOTE: ここのdeepcopyめっちゃ遅い
        new_label.data = [deepcopy(ol) for ol in full_label]
        # 前の音節とノートに関する処理
        for ol in new_label[1:]:
            if ol.previous_syllable[0].identity in ('pau', 'sil'):
                ol.a[0:3] = ['xx'] * 3
            if ol.previous_note.is_rest():
                ol.d[0:3] = ['xx'] * 3
                ol.d[3:8] = ['xx'] * 5
        # 現在の音節とノートに関する処理
        for ol in new_label:
            if ol.note.is_rest():
                ol.e[0:2] = ['xx'] * 2
        # 次の音節とノートに関する処理
        for ol in new_label[:-1]:
            if ol.next_syllable[0].identity in ('pau', 'sil'):
                ol.c[0:3] = ['xx'] * 3
            if ol.next_note.is_rest():
                ol.f[0:3] = ['xx'] * 3
                ol.f[3:8] = ['xx'] * 5

    # deepcopyが浅く処理が速い。ただし、Sinsyの出力と差異が生じてしまう。
    else:
        new_label = deepcopy(full_label)
        for ol in new_label:
            if ol.note.is_rest():
                ol.b[0:3] = ['xx'] * 3
                ol.e[0:3] = ['xx'] * 3

    return new_label


def main():
    """
    直接実行された時にやるやつ
    """
    path_hts_in = input('path_hts_in: ')
    label = load(path_hts_in)
    label.song.autofill()
    label.song.reset_time()
    path_hts_out = path_hts_in.replace('.lab', '_utaupy.lab')
    label.write(path_hts_out, strict_sinsy_style=True)


if __name__ == '__main__':
    main()



#! /usr/bin/env python3
# coding: utf-8
# Copyright (c) oatsu
"""
UTAUのプラグイン用のモジュール
utaupy.ust.Ust をもとに、ファイル入出力機能を変更したもの。
"""

from copy import deepcopy
# from pprint import pprint
from sys import argv
from typing import Callable

from utaupy import ust as _ust


def run(your_function: Callable, option=None, path=None):
    """
    UTAUプラグインスクリプトファイルの入出力をする。
    your_function: 実行したい関数
    arguments: 実行オプションとか
    path: UTAUから出力されるプラグインスクリプトのパス
    """
    if path is None:
        path = argv[1]
    # up.utauplugin.Plugin オブジェクトとしてプラグインスクリプトを読み取る
    plugin = load(path)
    # 目的のノート処理を実行
    if option is None:
        your_function(plugin)
    else:
        your_function(plugin, option)
    # プラグインスクリプトを上書き
    plugin.write(path)


def load(path: str, encoding='cp932'):
    """
    UTAUプラグイン一時ファイルを読み取る
    USTのやつを一部改変
    """
    # UtauPluginオブジェクト化
    plugin = UtauPlugin()
    plugin.load(path, encoding=encoding)
    return plugin


class UtauPlugin(_ust.Ust):
    """
    UTAUプラグインの一時ファイル用のクラス
    UST用のクラスを継承
    """
    def __init__(self):
        super().__init__()
        # プラグインのときは[#TRACKEND]が不要
        self.trackend = None

    def write(self, path: str, mode: str = 'w', encoding: str = 'cp932') -> str:
        """
        USTをファイル出力
        """
        # 文字列にする
        s = str(deepcopy(self)) + '\n'
        # ファイル出力
        with open(path, mode=mode, encoding=encoding) as f:
            f.write(s)
        return s



#! /usr/bin/env python3
# coding: utf-8
# Copyright (c) 2020 oatsu
"""
HTSフルコンテキストラベルをJSONと相互変換する。
"""

import re


def _load_hts_lines(lines: list) -> dict:
    """
    文字列のリスト(行のリスト)をもとに値を登録する。
    """
    # ラベル情報のリスト
    labels = []
    # HTSフルコンテキストラベルの各行を読み取っていく
    for line in lines:
        line_split = line.split(maxsplit=2)
        # 1行分の辞書
        d_line = {}
        # 時刻の情報 [発声開始時刻, 発声終了時刻]
        d_line['time'] = list(map(int, line_split[0:2]))
        # コンテキスト部分を取り出す
        str_contexts = line_split[2]
        # コンテキスト文字列を /A: などの文字列で区切って一次元リストにする
        l_contexts = re.split('/.:', str_contexts)
        # 特定の文字でさらに区切って二次元リストにする
        delimiters = re.escape('=+-~∼!@#$%^ˆ&;_|[]')
        l_contexts_2d = [re.split((f'[{delimiters}]'), s) for s in l_contexts]
        # 各種コンテキストを辞書に登録する
        d_line.update(dict(
            zip(('p', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j'), l_contexts_2d)))
        # ここからzipのもとのコード----------------------
        # d_line['p'] = l_contexts_2d[0]
        # d_line['a'] = l_contexts_2d[1]
        # d_line['b'] = l_contexts_2d[2]
        # d_line['c'] = l_contexts_2d[3]
        # d_line['d'] = l_contexts_2d[4]
        # d_line['e'] = l_contexts_2d[5]
        # d_line['f'] = l_contexts_2d[6]
        # d_line['g'] = l_contexts_2d[7]
        # d_line['h'] = l_contexts_2d[8]
        # d_line['i'] = l_contexts_2d[9]
        # d_line['j'] = l_contexts_2d[10]
        # ここまでzipのもとのコード----------------------

        # ラベル情報のリストに追加
        labels.append(d_line)
    return {'labels': labels}


def _load(path: str, encoding='utf-8') -> dict:
    """
    HTSフルコンテキストラベル(Sinsy用)のファイルを読み取り、辞書にする。
    """
    # ファイルを読み取って行のリストにする
    try:
        with open(path, mode='r', encoding=encoding) as f:
            lines = [line.rstrip('\r\n') for line in f.readlines()]
    except UnicodeDecodeError:
        with open(path, mode='r', encoding='cp932') as f:
            lines = [line.rstrip('\r\n') for line in f.readlines()]
    return _load_hts_lines(lines)


def _export_flatjson(d: dict, path) -> str:
    """
    JSON文字列でファイル出力する。
    1音素1行
    """
    s = ',\n'.join([f'        {str(d_line)}' for d_line in d['labels']])
    s = s.replace('\'', '"').replace('{', '{ ').replace('}', ' }')
    s = '{\n    \"labels\": [\n' + s + '\n    ]\n}\n'
    with open(path, mode='w', encoding='utf-8', newline='\n') as f:
        f.write(s)
    return s


def hts2json(path_lab, path_json):
    """
    HTSフルコンテキストラベルファイル(.lab) を
    JSONファイル(.json) に変換する。
    """
    _export_flatjson(_load(path_lab), path_json)


def main():
    """
    直接起動したときの動作。
    1つのラベルファイルをJSONに変換する。
    """
    path_lab = input('path_lab: ')
    path_json = path_lab.replace('.lab', '.json')
    hts2json(path_lab, path_json)


if __name__ == '__main__':
    main()



#! /usr/bin/env python3
# coding: utf-8
"""
日本語とアルファベットの対応表を扱うモジュールです。
"""


def main():
    """呼び出されても特に何もしない"""
    print('平仮名とアルファベットの対応表を扱うモジュールです。')


def load(path, encoding='utf-8') -> dict:
    """テーブルを読み取ってインスタンス生成"""
    path = str(path).strip('\'"')
    if path.endswith('.table'):
        return load_table_file(path, encoding=encoding)
    if path.endswith('.conf'):
        return load_conf_file(path, encoding=encoding)
    raise ValueError(f'Input path must end with ".table" or ".conf".: {path}')


def load_table_file(path_table, encoding='utf-8') -> dict:
    """テーブルを読み取ってインスタンス生成"""
    # ファイル読み取り
    try:
        with open(path_table, mode='r', encoding=encoding) as f:
            lines = [line.strip() for line in f.readlines()]
    except UnicodeDecodeError:
        try:
            with open(path_table, mode='r', encoding='cp932') as f:
                lines = [line.strip() for line in f.readlines()]
        except UnicodeDecodeError:
            with open(path_table, mode='r', encoding='utf-8') as f:
                lines = [line.strip() for line in f.readlines()]
    # 辞書にする
    d_table = {}
    for line in lines:
        line_split = line.split()
        d_table[line_split[0]] = line_split[1:]
    return d_table


def load_conf_file(path_conf, encoding='utf-8') -> dict:
    """音素分類用のファイルを読み取って辞書を返す"""
    # ファイル読み取り
    try:
        with open(path_conf, mode='r', encoding=encoding) as f:
            lines = [line.strip() for line in f.readlines()]
    except UnicodeDecodeError:
        try:
            with open(path_conf, mode='r', encoding='cp932') as f:
                lines = [line.strip() for line in f.readlines()]
        except UnicodeDecodeError:
            with open(path_conf, mode='r', encoding='utf-8') as f:
                lines = [line.strip() for line in f.readlines()]
    # 辞書にする
    d_conf = {'SILENCES': ['sil'], 'PAUSES': ['pau'], 'BREAKS': ['br']}
    for line in lines:
        key, value = line.split('=', 1)
        phonemes = value.strip('"').split(',')
        d_conf[key] = phonemes
    if 'PHONEME_CL' in d_conf:
        d_conf['BREAKS'] += d_conf['PHONEME_CL']

    return d_conf


if __name__ == '__main__':
    main()



#! /usr/bin/env python3
# coding: utf-8
"""
setParam用のINIファイルとデータを扱うモジュールです。
"""
import re
from collections import UserList

# TODO: setParam用のコメントファイルを扱えるようにする。

def main():
    """
    直接実行されたときの挙動
    """
    print('耳ロボPとsetParamに卍感謝卍')


def load(path, mode='r', encoding='cp932'):
    """
    otoiniを読み取ってオブジェクト生成
    """
    # otoiniファイルを読み取る
    path = path.strip('"')
    with open(path, mode=mode, encoding=encoding) as f:
        lines = [line.strip() for line in f.readlines()]

    # Otoクラスオブジェクトのリストを作る
    otoini = OtoIni()
    for line in lines:
        params = re.split('[=,]', line.strip())
        params = params[:2] + [float(v) for v in params[2:]]
        oto = Oto()
        (oto.filename, oto.alias, oto.offset, oto.consonant,
         oto.cutoff, oto.preutterance, oto.overlap) = params
        otoini.append(oto)
    return otoini


class OtoIni(UserList):
    """
    oto.iniを想定したクラス
    """

    def replace_aliases(self, before, after):
        """
        エイリアスを置換する
        """
        for oto in self:
            oto.alias = oto.alias.replace(before, after)
        return self

    def is_mono(self):
        """
        モノフォン形式のエイリアスになっているか判定する。
        すべてのエイリアスに空白がなければモノフォンと判断する。
        返り値はbool。
        """
        return all((' ' not in oto.alias) for oto in self)

    def monophonize(self):
        """
        音素ごとに分割する。
        otoini→label 変換の用途を想定
        音素の発声開始位置: 左ブランク=先行発声
        """
        # 新規OtoIniを作るために、otoを入れるリスト
        mono_otoini = OtoIni()
        for oto in self:
            phonemes = oto.alias.split()
            if len(phonemes) == 1:
                mono_otoini.append(oto)
            elif len(phonemes) in [2, 3]:
                name_wav = oto.filename
                # 1文字目(オーバーラップから先行発声まで)------------
                mono_oto = Oto()
                mono_oto.filename = name_wav
                mono_oto.alias = phonemes[0]
                mono_oto.offset = oto.offset + oto.overlap  # オーバーラップの位置に左ブランクを移動
                mono_oto.preutterance = 0
                mono_otoini.append(mono_oto)
                # 2文字目(先行発声から固定範囲まで)----------------
                mono_oto = Oto()
                mono_oto.filename = name_wav
                mono_oto.alias = phonemes[1]
                mono_oto.offset = oto.offset + oto.preutterance  # 先行発声の位置に左ブランクを移動
                mono_oto.preutterance = 0
                mono_otoini.append(mono_oto)
                if len(phonemes) == 3:
                    # 3文字目(固定範囲から右ブランクまで)----------------
                    mono_oto = Oto()
                    mono_oto.filename = name_wav
                    mono_oto.alias = phonemes[2]
                    mono_oto.offset = oto.offset + oto.consonant  # 固定範囲の位置に左ブランクを移動
                    mono_oto.preutterance = 0
                    mono_otoini.append(mono_oto)
            else:
                print('\n[ERROR in otoini.monophonize()]----------------')
                print('  1エイリアスの音素数は 1, 2, 3 以外対応していません。')
                print('  phonemes: {}'.format(phonemes))
                print('  文字を連結して処理を続行します。')
                print('-----------------------------------------------\n')
                mono_otoini.append(oto)
        return mono_otoini

    def write(self, path, mode='w', encoding='cp932'):
        """
        ファイル出力
        """
        s = '\n'.join([str(oto) for oto in self]) + '\n'
        with open(path, mode=mode, encoding=encoding) as f:
            f.write(s)
        return s


class Oto:
    """
    oto.ini中の1モーラ
    """

    def __init__(self):
        self.filename: str = ''
        self.alias: str = ''
        self.offset = 0
        self.consonant = 0
        self.cutoff = 0
        self.preutterance = 0
        self.overlap = 0
        self.comment = None

    def __str__(self):
        s = '{}={},{},{},{},{},{}'.format(
            self.filename,
            self.alias,
            round(float(self.offset), 4),
            round(float(self.consonant), 4),
            round(float(self.cutoff), 4),
            round(float(self.preutterance), 4),
            round(float(self.overlap), 4)
        )
        return s

    @property
    def cutoff2(self):
        """
        右ブランクを絶対時刻で取得する
        """
        cutoff = self.cutoff
        if cutoff > 0:
            raise ValueError(f'Cutoff(右ブランク) must be negative : {str(self)}')
        return self.offset - cutoff

    @cutoff2.setter
    def cutoff2(self, absolute_cutoff_time):
        """
        右ブランクを絶対時刻で受け取り、負の値で上書きする
        """
        if absolute_cutoff_time < 0:
            raise ValueError(
                f'Argument "absolute_cutoff_time" must be positive : {absolute_cutoff_time}')
        self.cutoff = self.offset - absolute_cutoff_time


if __name__ == '__main__':
    main()

if __name__ == '__init__':
    pass



#! /usr/bin/env python3
# coding: utf-8
# Copyright (c) 2020 oatsu
"""
setParamプラグイン用モジュール
## リンク
- setParam: http://nwp8861.web.fc2.com/soft/setParam
- setParamプラグインの入出力ファイルの仕様: http://nwp8861.blog92.fc2.com/blog-entry-365.html
"""

from . import otoini


def run(your_function):
    """
    setParamプラグイン用ファイルの入出力をする。
    """
    with open('inparam.txt', 'r') as f_ip:
        lines = [line.rstrip('\n') for line in f_ip.readlines]
        # プラグイン入出力用ファイル 'oto-autoEstimation.ini' のパス
        path_oto_autoestimation_ini = lines[1]
    # utaupy.otoini.OtoIni クラスインスタンスにする
    oto_autoestimation_ini = otoini.load(path_oto_autoestimation_ini)
    # やりたい処理を実行
    your_function(oto_autoestimation_ini)
    # 上書き保存
    oto_autoestimation_ini.write(path_oto_autoestimation_ini)



#! /usr/bin/env python3
# coding: utf-8
"""
USTファイルとデータを扱うモジュールです。
"""
import re
from collections import UserDict
from copy import deepcopy
from typing import List

from .utau import utau_appdata_root, utau_root

NOTENUM_TO_NOTENAME_DICT = {
    12: 'C0', 13: 'Db0', 14: 'D0', 15: 'Eb0', 16: 'E0', 17: 'F0',
    18: 'Gb0', 19: 'G0', 20: 'Ab0', 21: 'A0', 22: 'Bb0', 23: 'B0',
    24: 'C1', 25: 'Db1', 26: 'D1', 27: 'Eb1', 28: 'E1', 29: 'F1',
    30: 'Gb1', 31: 'G1', 32: 'Ab1', 33: 'A1', 34: 'Bb1', 35: 'B1',
    36: 'C2', 37: 'Db2', 38: 'D2', 39: 'Eb2', 40: 'E2', 41: 'F2',
    42: 'Gb2', 43: 'G2', 44: 'Ab2', 45: 'A2', 46: 'Bb2', 47: 'B2',
    48: 'C3', 49: 'Db3', 50: 'D3', 51: 'Eb3', 52: 'E3', 53: 'F3',
    54: 'Gb3', 55: 'G3', 56: 'Ab3', 57: 'A3', 58: 'Bb3', 59: 'B3',
    60: 'C4', 61: 'Db4', 62: 'D4', 63: 'Eb4', 64: 'E4', 65: 'F4',
    66: 'Gb4', 67: 'G4', 68: 'Ab4', 69: 'A4', 70: 'Bb4', 71: 'B4',
    72: 'C5', 73: 'Db5', 74: 'D5', 75: 'Eb5', 76: 'E5', 77: 'F5',
    78: 'Gb5', 79: 'G5', 80: 'Ab5', 81: 'A5', 82: 'Bb5', 83: 'B5',
    84: 'C6', 85: 'Db6', 86: 'D6', 87: 'Eb6', 88: 'E6', 89: 'F6',
    90: 'Gb6', 91: 'G6', 92: 'Ab6', 93: 'A6', 94: 'Bb6', 95: 'B6',
    96: 'C7', 97: 'Db7', 98: 'D7', 99: 'Eb7', 100: 'E7', 101: 'F7',
    102: 'Gb7', 103: 'G7', 104: 'Ab7', 105: 'A7', 106: 'Bb7', 107: 'B7',
    108: 'C8', 109: 'Db8', 110: 'D8', 111: 'Eb8', 112: 'E8', 113: 'F8',
    114: 'Gb8', 115: 'G8', 116: 'Ab8', 117: 'A8', 118: 'Bb8', 119: 'B8',
    120: 'C9', 121: 'Db9', 122: 'D9', 123: 'Eb9', 124: 'E9', 125: 'F9',
    126: 'Gb9', 127: 'G9',
    '12': 'C0', '13': 'Db0', '14': 'D0', '15': 'Eb0', '16': 'E0', '17': 'F0',
    '18': 'Gb0', '19': 'G0', '20': 'Ab0', '21': 'A0', '22': 'Bb0', '23': 'B0',
    '24': 'C1', '25': 'Db1', '26': 'D1', '27': 'Eb1', '28': 'E1', '29': 'F1',
    '30': 'Gb1', '31': 'G1', '32': 'Ab1', '33': 'A1', '34': 'Bb1', '35': 'B1',
    '36': 'C2', '37': 'Db2', '38': 'D2', '39': 'Eb2', '40': 'E2', '41': 'F2',
    '42': 'Gb2', '43': 'G2', '44': 'Ab2', '45': 'A2', '46': 'Bb2', '47': 'B2',
    '48': 'C3', '49': 'Db3', '50': 'D3', '51': 'Eb3', '52': 'E3', '53': 'F3',
    '54': 'Gb3', '55': 'G3', '56': 'Ab3', '57': 'A3', '58': 'Bb3', '59': 'B3',
    '60': 'C4', '61': 'Db4', '62': 'D4', '63': 'Eb4', '64': 'E4', '65': 'F4',
    '66': 'Gb4', '67': 'G4', '68': 'Ab4', '69': 'A4', '70': 'Bb4', '71': 'B4',
    '72': 'C5', '73': 'Db5', '74': 'D5', '75': 'Eb5', '76': 'E5', '77': 'F5',
    '78': 'Gb5', '79': 'G5', '80': 'Ab5', '81': 'A5', '82': 'Bb5', '83': 'B5',
    '84': 'C6', '85': 'Db6', '86': 'D6', '87': 'Eb6', '88': 'E6', '89': 'F6',
    '90': 'Gb6', '91': 'G6', '92': 'Ab6', '93': 'A6', '94': 'Bb6', '95': 'B6',
    '96': 'C7', '97': 'Db7', '98': 'D7', '99': 'Eb7', '100': 'E7', '101': 'F7',
    '102': 'Gb7', '103': 'G7', '104': 'Ab7', '105': 'A7', '106': 'Bb7', '107': 'B7',
    '108': 'C8', '109': 'Db8', '110': 'D8', '111': 'Eb8', '112': 'E8', '113': 'F8',
    '114': 'Gb8', '115': 'G8', '116': 'Ab8', '117': 'A8', '118': 'Bb8', '119': 'B8',
    '120': 'C9', '121': 'Db9', '122': 'D9', '123': 'Eb9', '124': 'E9', '125': 'F9',
    '126': 'Gb9', '127': 'G9'
}

NOTENAME_TO_NOTENUM_DICT = {
    'Cb0': 11, 'C0': 12, 'C#0': 13,
    'Db0': 13, 'D0': 14, 'D#0': 15,
    'Eb0': 15, 'E0': 16, 'E#0': 17,
    'Fb0': 16, 'F0': 17, 'F#0': 18,
    'Gb0': 18, 'G0': 19, 'G#0': 20,
    'Ab0': 20, 'A0': 21, 'A#0': 22,
    'Bb0': 22, 'B0': 23, 'B#0': 24,

    'Cb1': 23, 'C1': 24, 'C#1': 25,
    'Db1': 25, 'D1': 26, 'D#1': 27,
    'Eb1': 27, 'E1': 28, 'E#1': 29,
    'Fb1': 28, 'F1': 29, 'F#1': 30,
    'Gb1': 30, 'G1': 31, 'G#1': 32,
    'Ab1': 32, 'A1': 33, 'A#1': 34,
    'Bb1': 34, 'B1': 35, 'B#1': 36,

    'Cb2': 35, 'C2': 36, 'C#2': 37,
    'Db2': 37, 'D2': 38, 'D#2': 39,
    'Eb2': 39, 'E2': 40, 'E#2': 41,
    'Fb2': 40, 'F2': 41, 'F#2': 42,
    'Gb2': 42, 'G2': 43, 'G#2': 44,
    'Ab2': 44, 'A2': 45, 'A#2': 46,
    'Bb2': 46, 'B2': 47, 'B#2': 38,

    'Cb3': 47, 'C3': 48, 'C#3': 49,
    'Db3': 49, 'D3': 50, 'D#3': 51,
    'Eb3': 51, 'E3': 52, 'E#3': 53,
    'Fb3': 52, 'F3': 53, 'F#3': 54,
    'Gb3': 54, 'G3': 55, 'G#3': 56,
    'Ab3': 56, 'A3': 57, 'A#3': 58,
    'Bb3': 58, 'B3': 59, 'B#3': 60,

    'Cb4': 59, 'C4': 60, 'C#4': 61,
    'Db4': 61, 'D4': 62, 'D#4': 63,
    'Eb4': 63, 'E4': 64, 'E#4': 65,
    'Fb4': 64, 'F4': 65, 'F#4': 66,
    'Gb4': 66, 'G4': 67, 'G#4': 68,
    'Ab4': 68, 'A4': 69, 'A#4': 70,
    'Bb4': 70, 'B4': 71, 'B#4': 72,

    'Cb5': 71, 'C5': 72, 'C#5': 73,
    'Db5': 73, 'D5': 74, 'D#5': 75,
    'Eb5': 75, 'E5': 76, 'E#5': 77,
    'Fb5': 76, 'F5': 77, 'F#5': 78,
    'Gb5': 78, 'G5': 79, 'G#5': 79,
    'Ab5': 80, 'A5': 81, 'A#5': 82,
    'Bb5': 82, 'B5': 83, 'B#5': 84,

    'Cb6': 83, 'C6': 84, 'C#6': 85,
    'Db6': 85, 'D6': 86, 'D#6': 87,
    'Eb6': 87, 'E6': 88, 'E#6': 89,
    'Fb6': 88, 'F6': 89, 'F#6': 90,
    'Gb6': 90, 'G6': 91, 'G#6': 92,
    'Ab6': 92, 'A6': 93, 'A#6': 94,
    'Bb6': 94, 'B6': 95, 'B#6': 96,

    'Cb7': 95, 'C7': 96, 'C#7': 97,
    'Db7': 97, 'D7': 98, 'D#7': 99,
    'Eb7': 99, 'E7': 100, 'E#7': 101,
    'Fb7': 100, 'F7': 101, 'F#7': 102,
    'Gb7': 102, 'G7': 103, 'G#7': 104,
    'Ab7': 104, 'A7': 105, 'A#7': 106,
    'Bb7': 106, 'B7': 107, 'B#7': 108,

    'Cb8': 107, 'C8': 108, 'C#8': 109,
    'Db8': 109, 'D8': 110, 'D#8': 111,
    'Eb8': 111, 'E8': 112, 'E#8': 113,
    'Fb8': 112, 'F8': 113, 'F#8': 114,
    'Gb8': 114, 'G8': 115, 'G#8': 116,
    'Ab8': 116, 'A8': 117, 'A#8': 118,
    'Bb8': 118, 'B8': 119, 'B#8': 120,

    'Cb9': 119, 'C9': 120, 'C#9': 121,
    'Db9': 121, 'D9': 122, 'D#9': 123,
    'Eb9': 123, 'E9': 124, 'E#9': 125,
    'Fb9': 124, 'F9': 125, 'F#9': 126,
    'Gb9': 126, 'G9': 127
}


def notenum_as_abc(notenum) -> str:
    """
    音階番号をABC表記に変更する(C1=24, C4=60)
    """
    # raise DeprecationWarning(
    #     '関数 utaupy.ust.notenum_as_abc() は utaupy.utils.notenum2notename に名称変更しました。')
    return NOTENUM_TO_NOTENAME_DICT[int(notenum)]


def load(path: str, encoding: str = 'cp932'):
    """
    USTを読み取り
    """
    new_ust = Ust()
    new_ust.load(path.strip('"'), encoding=encoding)
    return new_ust


class Ust:
    """
    UST (UTAU Sequence Text) ファイルを扱うためのクラス
    """

    def __init__(self):
        super().__init__()
        # ノート(クラスオブジェクト)からなるリスト
        self.version = None   # [#VERSION]
        self.notes: List[Note] = []  # [#1234], [#INSERT], [#DELETE]
        self.setting = Note(tag='[#SETTING]')  # [#SETTING]
        self.setting['Tempo'] = 120
        del self.setting['Length']
        del self.setting['NoteNum']
        self.trackend = Note(tag='[#TRACKEND]')  # [#TRACKEND]
        self.next_note = None  # [#NEXT]
        self.previous_note = None  # [#PREV]

    def __str__(self):
        # self.notesが増減するので複製したものを扱う
        duplicated_self = deepcopy(self)
        # 特殊ノートを処理
        if len(self.setting) >= 2:
            duplicated_self.notes.insert(0, self.setting)
        if self.previous_note is not None:
            duplicated_self.notes.insert(0, self.previous_note)
        if self.next_note is not None:
            duplicated_self.notes.append(self.next_note)
        if self.trackend is not None:
            duplicated_self.notes.append(self.trackend)
        # 通常ノートを文字列にする
        s = '\n'.join(str(note) for note in duplicated_self.notes)
        # バージョン情報があれば先頭に追記
        if self.version is not None:
            str_version = f'[#VERSION]\nUST Version {str(duplicated_self.version)}'
            s = '\n'.join((str_version, s))
        return s

    def load(self, path: str, encoding='cp932'):
        """
        ファイルからインスタンス生成
        """
        # USTを文字列として取得
        try:
            with open(path, mode='r', encoding=encoding) as f:
                s = f.read().strip()
        except UnicodeDecodeError:
            try:
                with open(path, mode='r', encoding='utf-8') as f:
                    s = f.read().strip()
            except UnicodeDecodeError:
                with open(path, mode='r', encoding='utf-8_sig') as f:
                    s = f.read().strip()

        # USTの文字列をノート単位に分割
        l: List[str] = [f'[#{v.strip()}' for v in s.split('[#')][1:]
        l_2d = [s.split('\n') for s in l]

        # ノートのリストを作る
        for lines in l_2d:
            # 1行目: ノートの種類
            tag = lines[0]
            # print('reading entry:', tag)  # デバッグ用出力
            note = Note(tag=tag)
            # どこに登録するか決める
            if tag not in ('[#VERSION]', '[#SETTING]', '[#TRACKEND]', '[#PREV]', '[#NEXT]'):
                self.notes.append(note)
            elif tag == '[#VERSION]':
                self.version = lines[1].replace('UST Version ', '')
                continue
            elif tag == '[#SETTING]':
                self.setting = note
            elif tag == '[#PREV]':
                self.previous_note = note
            elif tag == '[#NEXT]':
                self.next_note = note
            elif tag == '[#TRACKEND]':
                self.trackend = note
            else:
                raise Exception('想定外のエラーです。開発者に連絡してください。:', tag, str(note))
            # 2行目移行: タグ以外の情報
            for line in lines[1:]:
                key, value = line.split('=', maxsplit=1)
                note[key] = value

        # 隠しパラメータ alternative_tempo を全ノートに設定
        self.reload_tempo()
        return self

    @property
    def tempo(self) -> float:
        """
        USTのグローバルテンポ
        """
        # 2020-12-19以前の実装------------
        # global_tempo = self.setting.get('Tempo', self.notes[0].tempo)
        # --------------------------------
        if 'Tempo' not in self.setting:
            self.setting['Tempo'] = self.notes[0].tempo
        return self.setting['Tempo']

    @tempo.setter
    def tempo(self, tempo):
        self.setting.tempo = tempo
        # self.notes[0].tempo = tempo
        self.reload_tempo()

    @property
    def voicedir(self):
        """
        UTAU音源のフォルダのパスを返す
        """
        voicedir = self.setting['VoiceDir']
        voicedir = voicedir.replace('%VOICE%', f'{utau_root()}\\voice')
        voicedir = voicedir.replace('%DATA%', utau_appdata_root())
        return voicedir

    @voicedir.setter
    def voicedir(self, path):
        self.setting['VoiceDir'] = path.strip('\'"')

    def clean_tempo(self):
        """
        ローカルテンポが不要な部分を削除する。
        """
        # まずはグローバルテンポを取得
        current_tempo = self.tempo
        # 不要なデータを削除
        for note in self.notes:
            if 'Tempo' in note:
                if note['Tempo'] == current_tempo:
                    del note['Tempo']
            else:
                current_tempo = note.tempo

    def reload_tempo(self):
        """
        1. グローバルテンポを1ノート目のテンポで上書きする。
        2. 各ノートでBPMが取得できるように note.alternative_tempo を全ノートに仕込む。
        """
        # ノートがないときは何もしない
        if len(self.notes) == 0:
            return
        # ここからはノートが1つ以上あるとき
        if 'Tempo' in self.notes[0]:
            self.setting['Tempo'] = self.notes[0]['Tempo']
        current_tempo = self.setting['Tempo']

        # [#PREV]にalternative_tempoを登録
        previous_note = self.previous_note
        if previous_note is not None:
            previous_note.alternative_tempo = previous_note.get('Tempo', current_tempo)

        # 通常のノートにalternative_tempoを登録
        for note in self.notes:
            if 'Tempo' in note:
                if float(note['Tempo']) == current_tempo:
                    del note['Tempo']
                else:
                    current_tempo = note['Tempo']
            # current_tempo = note.get('Tempo', current_tempo)
            note.alternative_tempo = float(current_tempo)

        # [#NEXT]にalternative_tempoを登録
        next_note = self.next_note
        if next_note is not None:
            next_note.alternative_tempo = next_note.get('Tempo', current_tempo)

        self.clean_tempo()

    def reload_tag_number(self, start: int = 0):
        """
        start: 開始番号 (0なら[#0000]から)
        各ノートのノート番号を振りなおす。
        ファイル出力時に実行することを想定。
        """
        for i, note in enumerate(self.notes, start):
            note.tag = f'[#{str(i).zfill(4)}]'

    # ノート一括編集系関数ここから----------------------------------------------
    def replace_lyrics(self, before: str, after: str):
        """
        歌詞を一括置換（文字列指定・破壊的処理）
        """
        for note in self.notes:
            note.lyric = note.lyric.replace(before, after)

    def translate_lyrics(self, before, after):
        """
        歌詞を一括置換（複数文字指定・破壊的処理）
        """
        for note in self.notes:
            note.lyric = note.lyric.translate(before, after)

    def vcv2cv(self):
        """
        歌詞を平仮名連続音から単独音にする
        """
        for note in self.notes:
            note.lyric = note.lyric.split(' ', 1)[-1]
    # ノート一括編集系関数ここまで----------------------------------------------

    def insert_note(self, i: int):
        """
        i 番目の区切りに新規ノートを挿入する。
        このときの i は音符のみのインデックス。
        編集するために、挿入したノートを返す。
        """
        note = Note()
        # note.tag = '[#INSERT]'
        self.notes.insert(i, note)
        return note

    def delete_note(self, i: int):
        """
        i 番目のノートを [#DELETE] する。
        """
        self.notes[i].tag = '[#DELETE]'

    def make_finalnote_R(self):
        """
        Ustの最後のノートが休符 になるようにする
        """
        last_note = self.notes[-1]
        # Ust内の最後はTRACKENDなので後ろから2番目のノートで判定
        if last_note.lyric not in ('pau', 'sil', 'R'):
            rest_note = deepcopy(last_note)
            rest_note.lyric = 'R'
            rest_note.length = 480
            self.notes.append(rest_note)
        self.reload_tempo()

    def write(self, path: str, mode: str = 'w', encoding: str = 'cp932') -> str:
        """
        USTをファイル出力
        """
        duplicated_self = deepcopy(self)
        # [#DELETE] なノートをファイル出力しないために削除
        duplicated_self.notes = [
            note for note in duplicated_self.notes if note.tag != '[#DELETE]'
        ]
        # テンポを整理する
        duplicated_self.reload_tempo()
        # ノート番号を振りなおす
        duplicated_self.reload_tag_number()
        # 文字列にする
        s = str(duplicated_self) + '\n'
        # ファイル出力
        with open(path, mode=mode, encoding=encoding) as f:
            f.write(s)
        return s


class Note(UserDict):
    """
    UST内のノート
    """

    def __init__(self, tag: str = '[#INSERT]'):
        super().__init__()
        self['Tag'] = tag
        self.alternative_tempo = None
        self.length = 480
        self.notenum = 60

    def __str__(self):
        lines = [self['Tag']] + [f'{k}={v}' for (k, v) in self.items() if k != 'Tag']
        return '\n'.join(lines)

    @property
    def tag(self) -> str:
        """
        ノート識別用のタグ
        """
        return self['Tag']

    @tag.setter
    def tag(self, s):
        self['Tag'] = s

    @property
    def length(self) -> int:
        """
        ノート長[Ticks]
        """
        return int(self['Length'])

    @length.setter
    def length(self, x):
        self['Length'] = str(x)

    @property
    def length_ms(self) -> float:
        """
        ノート長[ms]
        """
        return 125 * float(self['Length']) / self.tempo

    @length_ms.setter
    def length_ms(self, x):
        self['Length'] = str(round(x * self.tempo / 125))

    @property
    def lyric(self) -> str:
        """
        歌詞
        """
        return self['Lyric']

    @lyric.setter
    def lyric(self, x):
        self['Lyric'] = x

    @property
    def notenum(self) -> int:
        """
        音階番号
        """
        return int(self['NoteNum'])

    @notenum.setter
    def notenum(self, x):
        self['NoteNum'] = str(x)

    @property
    def notename(self) -> str:
        """
        音名（C4などの表記）
        """
        return NOTENUM_TO_NOTENAME_DICT[self.notenum]

    @notename.setter
    def notename(self, notename: str):
        self.notenum = NOTENAME_TO_NOTENUM_DICT[str(notename)]

    @property
    def tempo(self) -> float:
        """
        ローカルBPM
        """
        return float(self.get('Tempo', self.alternative_tempo))

    @tempo.setter
    def tempo(self, x):
        self['Tempo'] = str(x)

    @property
    def pbs(self) -> list:
        """
        PBS (mode2ピッチ開始位置[ms])
        例) PBS=-104;20.0
        """
        # 辞書には文字列で登録してある
        str_pbs = self['PBS']
        # 浮動小数のリストに変換
        list_pbs = list(map(float, re.split('[;,]', str_pbs)))
        # PBSの値をリストで返す
        return list_pbs

    @pbs.setter
    def pbs(self, list_pbs):
        s1 = f'{int(list_pbs[0])};'
        s2 = ','.join(map(str, list_pbs[1:]))

        str_pbs = s1 + s2
        self['PBS'] = str_pbs

    @property
    def pbw(self) -> List[int]:
        """
        PBW (mode2ピッチ点の間隔[ms]) を取得
        例) PBW=77,163
        """
        # 辞書には文字列で登録してある
        s_pbw = self['PBW']
        # 整数のリストに変換
        l_pbw = list(map(int, s_pbw.split(',')))
        # PBWの値をリストで返す
        return l_pbw

    @pbw.setter
    def pbw(self, list_pbw: List[int]):
        # リストを整数の文字列に変換
        str_pbw = ','.join(list(map(str, map(int, list_pbw))))
        self['PBW'] = str_pbw

    @property
    def pby(self) -> list:
        """
        PBY (mode2ピッチ点の高さ) を取得
        例) PBY=-10.6,0.0
        """
        # 辞書には文字列で登録してある
        s_pby = self['PBY']
        # 整数のリストに変換
        l_pby = list(map(float, s_pby.split(',')))
        # PBYの値をリストで返す
        return l_pby

    @pby.setter
    def pby(self, list_pby):
        # リストを小数の文字列に変換
        str_pby = ','.join(list(map(str, map(float, list_pby))))
        self['PBY'] = str_pby

    @property
    def pbm(self) -> List[str]:
        """
        PBM (mode2ピッチ点の形状) を取得
        例) PBY=,,,,
        """
        # 辞書には文字列で登録してある
        s_pby = self['PBM']
        # 整数のリストに変換
        l_pbm = s_pby.split(',')
        # PBYの値をリストで返す
        return l_pbm

    @pbm.setter
    def pbm(self, list_pbm: list):
        # リストを文字列に変換
        str_pbm = ','.join(list_pbm)
        self['PBM'] = str_pbm

    @property
    def velocity(self) -> int:
        """
        子音速度
        """
        return int(self.get('Velocity', 100))

    @velocity.setter
    def velocity(self, x: int):
        self['Velocity'] = int(x)

    @property
    def flags(self) -> str:
        """
        フラグ
        """
        return self.get('Flags', '')

    @flags.setter
    def flags(self, flags: str):
        self['Flags'] = str(flags)

    @property
    def label(self) -> str:
        """
        ノートのラベル
        UTAUエディタのピアノロール画面で、ローカルテンポの上に表示されるやつ
        """
        return self.get('Label', '')

    @label.setter
    def label(self, label: str):
        self['Label'] = str(label)

    # ここからノート操作系-----------------------------------------------------

    def delete(self):
        """選択ノートを削除"""
        self.tag = '[#DELETE]'
        return self

    # def insert(self):
    #     """ノートを挿入(したい)"""
    #     self.tag = '[#INSERT]'
    #     return self

    def refresh(self):
        """
        ノートの情報を引き継ぎつつ、自由にいじれるようにする
        UTAUプラグインは値の上書きはできるが削除はできない。
        一旦ノートを削除して新規ノートとして扱う必要がある。
        """
        self.tag = '[#DELETE]\n[#INSERT]'

    def suppin(self):
        """
        ノートの情報を最小限にする
        """
        new_data = {}
        new_data['Tag'] = '[#DELETE]\n[#INSERT]'
        new_data['Lyric'] = self.lyric
        new_data['Length'] = self.length
        new_data['NoteNum'] = self.notenum
        self.data = new_data
    # ここまでノート操作系-----------------------------------------------------


def main():
    """
    実行されたときの挙動
    """
    print('デフォ子かわいいよデフォ子\n')

    print('ust読み取りテストをします。')
    path = input('ustのパスを入力してください。\n>>> ')
    ust = load(path)
    print(ust)

    input('\nPress Enter to exit.')


if __name__ == '__main__':
    main()



#! /usr/bin/env python3
# coding: utf-8
# Copyright (c) 2020 oatsu
"""
UTAU音源を扱うモジュール
"""
import platform
from collections import UserDict
from glob import glob
from os.path import dirname, expandvars

from . import otoini as _otoini

# Windowsのとき
if str(platform.system) == 'Windows':
    import winreg

    def utau_root() -> str:
        """
        utau.exe のフォルダのパスを返す。

        - 拡張子 '.ust' は UTAUSequenceText としてレジストリに登録されている。
        - UTAUSequenceText に関連付けられている UTAU.exeのパスを取得する。
        """
        # 現在のユーザーにおいて、ustファイルに関連付けられているコマンド取得して、
        # '"C:\\UTAU\\utau.exe" "%1"' のような文字列を得る
        reg_key = winreg.OpenKeyEx(winreg.HKEY_CURRENT_USER,
                                   r'Software\Classes\UTAUSequenceText\shell\open\command')
        reg_data, _ = winreg.QueryValueEx(reg_key, '')
        winreg.CloseKey(reg_key)
        # reg_data の前半部分だけ取り出して 'C:\\UTAU\\utau.exe' にする
        path_utau_exe = reg_data.split(r'" "')[0].strip('"')  #
        # 'C:\\UTAU' の形にして返す
        return dirname(path_utau_exe)
# Windowsじゃないとき
else:
    def utau_root() -> str:
        """
        utau.exe のフォルダのパスを返す。

        - 拡張子 '.ust' は UTAUSequenceText としてレジストリに登録されている。
        - UTAUSequenceText に関連付けられている UTAU.exeのパスを取得する。
        """
        # 'C:\\UTAU' の形にして返す
        return ''


def utau_appdata_root() -> str:
    r"""
    プラグインとか音源が置いてあるフォルダのパスを返す。
    C:\Users\{username}\AppData\Roaming\UTAU
    """
    return expandvars(r'%APPDATA%\UTAU')


class PrefixMap(UserDict):
    """
    UTAUの多音階音源用の prefixmap ファイルを扱うクラス。
    prefixmap_obj[音階名] または prefixmap_obj[音階番号] で
    サフィックス文字列を取得できるようにしたい。
    """

    def __init__(self, path):
        super().__init__()
        self.comment_lines = []
        with open(path, 'r') as prefixmap_file:
            lines = prefixmap_file.readlines()
        # 改行文字を削除、コメント行を無視
        lines = [line.rstrip('\r\n') for line in lines if not line.startswith('//')]
        # 空白で区切って {音程: suffix文字列} の辞書にする
        lines_2d = [line.split(maxsplit=1) for line in lines]
        # 音程表記をUSTでの音階番号に変換
        abc_to_notenum = {
            'C1': '24', 'C#1': '25', 'D1': '26', 'D#1': '27', 'E1': '28', 'F1': '29',
            'F#1': '30', 'G1': '31', 'G#1': '32', 'A1': '33', 'A#1': '34', 'B1': '35',
            'C2': '36', 'C#2': '37', 'D2': '38', 'D#2': '39', 'E2': '40', 'F2': '41',
            'F#2': '42', 'G2': '43', 'G#2': '44', 'A2': '45', 'A#2': '46', 'B2': '47',
            'C3': '48', 'C#3': '49', 'D3': '50', 'D#3': '51', 'E3': '52', 'F3': '53',
            'F#3': '54', 'G3': '55', 'G#3': '56', 'A3': '57', 'A#3': '58', 'B3': '59',
            'C4': '60', 'C#4': '61', 'D4': '62', 'D#4': '63', 'E4': '64', 'F4': '65',
            'F#4': '66', 'G4': '67', 'G#4': '68', 'A4': '69', 'A#4': '70', 'B4': '71',
            'C5': '72', 'C#5': '73', 'D5': '74', 'D#5': '75', 'E5': '76', 'F5': '77',
            'F#5': '78', 'G5': '79', 'G#5': '80', 'A5': '81', 'A#5': '82', 'B5': '83',
            'C6': '84', 'C#6': '85', 'D6': '86', 'D#6': '87', 'E6': '88', 'F6': '89',
            'F#6': '90', 'G6': '91', 'G#6': '92', 'A6': '93', 'A#6': '94', 'B6': '95',
            'C7': '96', 'C#7': '97', 'D7': '98', 'D#7': '99', 'E7': '100', 'F7': '101',
            'F#7': '102', 'G7': '103', 'G#7': '104', 'A7': '105', 'A#7': '106', 'B7': '107',
            'C8': '108', 'C#8': '109', 'D8': '110', 'D#8': '111', 'E8': '112', 'F8': '113',
            'F#8': '114', 'G8': '115', 'G#8': '116', 'A8': '117', 'A#8': '118', 'B8': '119'
        }
        # {音階番号: suffix文字列} の辞書になる
        for line in lines_2d:
            if len(line) == 1:
                self.data.update({abc_to_notenum[line[0]]: ''})
            else:
                self.data.update({abc_to_notenum[line[0]]: line[1]})


class UtauVoiceBank():
    """
    UTAU音源の原音設定を扱うクラス。
    音階と歌詞（連続音）を指定したら、原音設定を返すようにしたい。
    """

    def __init__(self, path):
        super().__init__()
        self.path = path
        self.prefixmap = PrefixMap(f'{self.path}/prefix.map')
        self.otoini = _otoini.OtoIni()
        # 原音設定ファイルを取得
        all_otoini_paths = glob(f'{path}/**/oto.ini', recursive=True)
        for path_otoini in all_otoini_paths:
            self.otoini.data += _otoini.load(path_otoini).data



    def autoselect_alias(self, utaupy_ust_note):
        """
        voicebank       : utaupy.utau.UtauVoiceBank オブジェクト
        utaupy_ust_note : utaupy.Ust.Note オブジェクト
        use_atalias     : Noteオブジェクトの '@alias' を使用するかどうか。
                          voicebankで指定しているUTAU音源が 対象USTのUTAU音源とは違う場合、
                          Falseにするほうがよい。

        Noteオブジェクトの歌詞と音程から、使用する原音に対応するエイリアスを決定する。
        ノート情報に @alias があればそのエイリアスを使用し、
        無い場合は prefix.map を使って決定する。
        """
        lyric = utaupy_ust_note.lyric
        if lyric == 'R':
            return 'R'
        # USTでエイリアスを強制指定している場合
        if lyric.startswith('?'):
            return lyric.lstrip('?')
        # 普通の歌詞の場合はprefixmapを参照してサフィックス追加
        # TODO: すでにサフィックスがある場合に不具合を回避する必要がある
        alias = lyric + self.prefixmap[str(utaupy_ust_note.notenum)]
        return alias


    def get_oto(self, utaupy_ust_note, suffix_exists=False):
        """
        voicebank       : utaupy.utau.UtauVoiceBank オブジェクト
        utaupy_ust_note : utaupy.Ust.Note オブジェクト
        use_atparam     : Noteオブジェクトの '@alias' などを使用するかどうか。
                          voicebankで指定しているUTAU音源が 対象USTのUTAU音源とは違う場合、
                          Falseにするほうがよい。
        Noteオブジェクトに対応する原音設定の値を取得する。
        """
        if suffix_exists is False:
            alias = self.autoselect_alias(utaupy_ust_note)
        # 原音にちゃんとあるかどうか
        if alias in self.otoini:
            oto_for_the_alias = self.otoini[alias]
        # なければ全部の数値がゼロの原音設定値を返す
        else:
            oto_for_the_alias = _otoini.Oto()
            oto_for_the_alias.alias = alias
        return oto_for_the_alias


    def autoadjust_parameters(self, utaupy_ust_ust):
        """
        voicebank       : utaupy.utau.UtauVoiceBank オブジェクト
        utaupy_ust_ust  : utaupy.Ust.Ust オブジェクト

        # TODO: プラグイン用に'@'パラメータに対応させる。
        UTAUの「パラメータ自動調整」機能に相当するメソッド。
        ・エイリアス(@alias)
        ・STP(StartPoint)
        ・オーバーラップ(VoiceOverlap)
        ・先行発声(PreUtterance)
        の4つの値を調整する。

        # TODO: 子音速度100以外に対応させる。
        仕様
        前のノートが休符な時は、原音値をそのまま入力する。
        前のノートが休符でない場合
            前のノートの長さの半分が、oto.overlapからoto.preutteranceまでの長さよりも短い場合
                差分をSTPで削る
        """
        notes = utaupy_ust_ust.notes
        for i, note in enumerate(notes[1:], 1):
            halflen = notes[i - 1].length_ms / 2
            oto = self.get_oto(note, suffix_exists=True)
            if halflen < oto.preutterance - oto.overlap:
                at_preuttr = halflen * oto.preutterance / (oto.preutterance - oto.overlap)
                at_overlap = halflen * oto.overlap / (oto.preutterance - oto.overlap)
                note['StartPoint'] = oto.preutterance - at_preuttr
                note['PreUtterance'] = at_preuttr
                note['VoiceOverlap'] = at_overlap
            else:
                note['StartPoint'] = 0
                note['PreUtterance'] = oto.preutterance
                note['VoiceOverlap'] = oto.overlap



#! /usr/bin/env python3
# coding: utf-8
# Copyright (c) 2020 oatsu
"""
USTファイルをHTSフルラベルに変換する。
Songオブジェクトを使って生成する。
いまのところ日本語にしか対応していないので注意。

対象
    CQS:
        p11, p12, p13, p14
        a1, a2, a3,
        b1, b2, b3,
        c1, c2, c3,
        d1, d2, d3, (d5), d6, d7, d8,
        e1, e2, e3, (e5), e6, e7, e8, e57, e58,
        f1, f2, f3, (f5), f6, f7, f8
    QS:
        p1, p3, p4, p5

d2, d3, e2, e3, f2, f3 はスケール判定が必要なため、実行時にキー(スケール)を指定する必要がある。
キーを指定しない場合は上記の値は'xx'のままにする。
d3, e3, f3 には 'xx' を代入する。歌うときに休符の学習データ引っ張ってきそうな気はする。
"""
import utaupy as up


def ustnote2htsnote(
        ust_note: up.ust.Note, d_table: dict, key_of_the_note: int = None) -> up.hts.Note:
    """
    utaupy.ust.Note を utaupy.hts.Note に変換する。
    """
    # ノート全体の情報を登録
    hts_note = up.hts.Note()
    # e1
    hts_note.notenum = ust_note.notenum
    # e2-e3
    if key_of_the_note is not None:
        # e2
        hts_note.relative_pitch = (ust_note.notenum - key_of_the_note) % 12
        # e3
        hts_note.key = key_of_the_note
    # e5
    hts_note.tempo = round(ust_note.tempo)
    # e8
    hts_note.length = round(ust_note.length / 20)

    # UST内のノートの歌詞を空白で区切る
    kana_lyrics = ust_note.lyric.replace('っ', ' っ ').split()

    # TODO: 音節数を2以上にできるようにする
    hts_syllable = up.hts.Syllable()
    # かな→ローマ字 で音素変換する
    phonemes = []
    for lyric in kana_lyrics:
        phonemes += d_table.get(lyric, [lyric])
    # 音素を追加していく
    for phoneme in phonemes:
        hts_phoneme = up.hts.Phoneme()
        # p4
        hts_phoneme.identity = phoneme
        # p9
        # ustのローカルフラグが設定されている時だけp9に記入
        if ust_note.flags != '':
            hts_phoneme.flag = ust_note.flags
        # 音節に追加
        hts_syllable.append(hts_phoneme)
    hts_note.append(hts_syllable)

    return hts_note


def ustobj2songobj(
        ust: up.ust.Ust, d_table: dict, key_of_the_note: int = None) -> up.hts.Song:
    """
    Ustオブジェクトをノートごとに処理して、HTS用に変換する。
    日本語歌詞を想定するため、音節数は1とする。促音に注意。

    ust: Ustオブジェクト
    d_table: 日本語→ローマ字変換テーブル

    key_of_the_note:
        曲のキーだが、USTからは判定できない。
        Sinsyでは 0 ~ 11 または 'xx' である。
    """
    song = up.hts.Song()
    ust_notes = ust.notes
    # Noteオブジェクトの種類を変換
    for ust_note in ust_notes:
        hts_note = ustnote2htsnote(ust_note, d_table, key_of_the_note=key_of_the_note)
        song.append(hts_note)

    # ノート長や位置などを自動補完
    song.autofill()
    # 発声開始時刻と終了時刻をノート長に応じて設定
    song.reset_time()
    return song


def ust2hts(
        path_ust: str, path_hts: str, path_table: str,
        strict_sinsy_style: bool = True, as_mono: bool = False):
    """
    USTファイルをLABファイルに変換する。
    """
    ust = up.ust.load(path_ust)
    d_table = up.table.load(path_table, encoding='utf-8')
    # Ust → HTSFullLabel
    hts_song = ustobj2songobj(ust, d_table)
    # HTSFullLabel中の重複データを削除して整理
    # ファイル出力
    hts_song.write(path_hts, strict_sinsy_style=strict_sinsy_style, as_mono=as_mono)


# def main():
#     """
#     USTファイルをLABファイルおよびJSONファイルに変換する。
#     """
#     # 各種パスを指定
#     path_table = input('path_table: ')
#     path_ust_in = input('path_ust: ')
#     path_hts_out = \
#         dirname(path_ust_in) + '/' + splitext(basename(path_ust_in))[0] + '_ust2hts.lab'
#     path_json_out = \
#         dirname(path_ust_in) + '/' + splitext(basename(path_ust_in))[0] + '_ust2hts.json'
#     # 変換
#     ust2hts(path_ust_in, path_hts_out, path_table, strict_sinsy_style=False)
#     # jsonファイルにも出力する。
#     hts2json(path_hts_out, path_json_out)
#
#
# if __name__ == '__main__':
#     main()
